from typing import Optional

from moto.core.exceptions import RESTError
from moto.core.responses import ActionResult, BaseResponse, EmptyResult
from moto.core.serialize import return_if_not_empty

from .exceptions import ListenerOrBalancerMissingError, TargetGroupNotFoundError
from .models import ELBv2Backend, elbv2_backends
from .ssl_policies import SSL_POLICIES, build_policy


def transform_dict(data: dict[str, str]) -> list[dict[str, str]]:
    transformed = [{"Key": key, "Value": value} for key, value in data.items()]
    return transformed


def transform_certificates(data: list[str]) -> Optional[list[dict[str, str]]]:
    return [{"CertificateArn": cert} for cert in data] or None


class ELBV2Response(BaseResponse):
    RESPONSE_KEY_PATH_TO_TRANSFORMER = {
        "CreateListenerOutput.Listeners.Listener.Certificates": transform_certificates,
        "CreateTargetGroupOutput.TargetGroups.TargetGroup.LoadBalancerArns": return_if_not_empty,
        "DescribeListenerAttributesOutput.Attributes": transform_dict,
        "DescribeListenersOutput.Listeners.Listener.Certificates": transform_certificates,
        "DescribeLoadBalancerAttributesOutput.Attributes": transform_dict,
        "DescribeTagsOutput.TagDescriptions.TagDescription.Tags": transform_dict,
        "DescribeTargetGroupAttributesOutput.Attributes": transform_dict,
        "ModifyListenerAttributesOutput.Attributes": transform_dict,
        "ModifyListenerOutput.Listeners.Listener.Certificates": transform_certificates,
        "ModifyLoadBalancerAttributesOutput.Attributes": transform_dict,
        "ModifyTargetGroupAttributesOutput.Attributes": transform_dict,
    }

    def __init__(self) -> None:
        super().__init__(service_name="elbv2")
        self.automated_parameter_parsing = True

    @property
    def elbv2_backend(self) -> ELBv2Backend:
        return elbv2_backends[self.current_account][self.region]

    def create_load_balancer(self) -> ActionResult:
        params = self._get_params()
        load_balancer_name = params.get("Name")
        subnet_ids = params.get("Subnets", [])
        subnet_mappings = params.get("SubnetMappings", [])
        security_groups = params.get("SecurityGroups", [])
        scheme = params.get("Scheme")
        loadbalancer_type = params.get("Type")
        ip_address_type = params.get("IpAddressType", "ipv4")
        tags = params.get("Tags")
        load_balancer = self.elbv2_backend.create_load_balancer(
            name=load_balancer_name,  # type: ignore
            security_groups=security_groups,
            subnet_ids=subnet_ids,
            subnet_mappings=subnet_mappings,
            scheme=scheme,  # type: ignore
            loadbalancer_type=loadbalancer_type,
            ip_address_type=ip_address_type,
            tags=tags,
        )
        result = {"LoadBalancers": [load_balancer]}
        return ActionResult(result)

    def create_rule(self) -> ActionResult:
        params = self._get_params()
        rule = self.elbv2_backend.create_rule(
            listener_arn=params["ListenerArn"],
            conditions=params["Conditions"],
            priority=params["Priority"],
            actions=params["Actions"],
            tags=params.get("Tags"),
        )
        result = {"Rules": [rule]}
        return ActionResult(result)

    def create_target_group(self) -> ActionResult:
        params = self._get_params()
        name = params.get("Name")
        vpc_id = params.get("VpcId")
        protocol = params.get("Protocol")
        protocol_version = params.get("ProtocolVersion")
        port = params.get("Port")
        healthcheck_protocol = self._get_param("HealthCheckProtocol")
        healthcheck_port = self._get_param("HealthCheckPort")
        healthcheck_path = self._get_param("HealthCheckPath")
        healthcheck_interval_seconds = self._get_int_param("HealthCheckIntervalSeconds")
        healthcheck_timeout_seconds = self._get_int_param("HealthCheckTimeoutSeconds")
        healthcheck_enabled = self._get_param("HealthCheckEnabled")
        healthy_threshold_count = self._get_param("HealthyThresholdCount")
        unhealthy_threshold_count = self._get_param("UnhealthyThresholdCount")
        matcher = params.get("Matcher")
        target_type = params.get("TargetType", "instance")
        ip_address_type = params.get("IpAddressType")
        tags = params.get("Tags")

        target_group = self.elbv2_backend.create_target_group(
            name,  # type: ignore
            vpc_id=vpc_id,
            protocol=protocol,
            protocol_version=protocol_version,
            port=port,
            healthcheck_protocol=healthcheck_protocol,
            healthcheck_port=healthcheck_port,
            healthcheck_path=healthcheck_path,
            healthcheck_interval_seconds=healthcheck_interval_seconds,
            healthcheck_timeout_seconds=healthcheck_timeout_seconds,
            healthcheck_enabled=healthcheck_enabled,
            healthy_threshold_count=healthy_threshold_count,
            unhealthy_threshold_count=unhealthy_threshold_count,
            matcher=matcher,
            ip_address_type=ip_address_type,
            target_type=target_type,
            tags=tags,
        )
        result = {"TargetGroups": [target_group]}
        return ActionResult(result)

    def create_listener(self) -> ActionResult:
        params = self._get_params()
        load_balancer_arn = self._get_param("LoadBalancerArn")
        protocol = self._get_param("Protocol")
        port = self._get_param("Port")
        ssl_policy = self._get_param("SslPolicy", "ELBSecurityPolicy-2016-08")
        certificates = self._get_param("Certificates", [])
        if certificates:
            certificate = certificates[0].get("CertificateArn")
        else:
            certificate = None
        default_actions = params.get("DefaultActions", [])
        alpn_policy = params.get("AlpnPolicy", [])
        tags = params.get("Tags")

        listener = self.elbv2_backend.create_listener(
            load_balancer_arn=load_balancer_arn,
            protocol=protocol,
            port=port,
            ssl_policy=ssl_policy,
            certificate=certificate,
            actions=default_actions,
            alpn_policy=alpn_policy,
            tags=tags,
        )

        result = {"Listeners": [listener]}
        return ActionResult(result)

    def describe_load_balancers(self) -> ActionResult:
        arns = self._get_param("LoadBalancerArns", [])
        names = self._get_param("Names", [])
        all_load_balancers = list(
            self.elbv2_backend.describe_load_balancers(arns, names)
        )
        marker = self._get_param("Marker")
        all_names = [balancer.name for balancer in all_load_balancers]
        if marker:
            start = all_names.index(marker) + 1
        else:
            start = 0
        page_size = self._get_int_param(
            "PageSize", 50
        )  # the default is 400, but using 50 to make testing easier
        load_balancers_resp = all_load_balancers[start : start + page_size]
        next_marker = None
        if len(all_load_balancers) > start + page_size:
            next_marker = load_balancers_resp[-1].name
        result = {"LoadBalancers": load_balancers_resp, "NextMarker": next_marker}
        return ActionResult(result)

    def describe_rules(self) -> ActionResult:
        listener_arn = self._get_param("ListenerArn")
        rule_arns = self._get_param("RuleArns")
        all_rules = list(self.elbv2_backend.describe_rules(listener_arn, rule_arns))
        all_arns = [rule.arn for rule in all_rules]
        page_size = self._get_int_param("PageSize", 50)  # set 50 for temporary

        marker = self._get_param("Marker")
        if marker:
            start = all_arns.index(marker) + 1
        else:
            start = 0
        rules_resp = all_rules[start : start + page_size]
        next_marker = None

        if len(all_rules) > start + page_size:
            next_marker = rules_resp[-1].arn
        result = {"Rules": rules_resp, "NextMarker": next_marker}
        return ActionResult(result)

    def describe_target_groups(self) -> ActionResult:
        load_balancer_arn = self._get_param("LoadBalancerArn")
        target_group_arns = self._get_param("TargetGroupArns", [])
        names = self._get_param("Names", [])
        target_groups = self.elbv2_backend.describe_target_groups(
            load_balancer_arn, target_group_arns, names
        )
        result = {"TargetGroups": target_groups}
        return ActionResult(result)

    def describe_target_group_attributes(self) -> ActionResult:
        target_group_arn = self._get_param("TargetGroupArn")
        target_group = self.elbv2_backend.target_groups.get(target_group_arn)
        if not target_group:
            raise TargetGroupNotFoundError()
        result = {"Attributes": target_group.attributes}
        return ActionResult(result)

    def describe_listeners(self) -> ActionResult:
        load_balancer_arn = self._get_param("LoadBalancerArn")
        listener_arns = self._get_param("ListenerArns", [])
        if not load_balancer_arn and not listener_arns:
            raise ListenerOrBalancerMissingError()

        listeners = self.elbv2_backend.describe_listeners(
            load_balancer_arn, listener_arns
        )
        result = {"Listeners": listeners}
        return ActionResult(result)

    def delete_load_balancer(self) -> ActionResult:
        arn = self._get_param("LoadBalancerArn")
        self.elbv2_backend.delete_load_balancer(arn)
        return EmptyResult()

    def delete_rule(self) -> ActionResult:
        arn = self._get_param("RuleArn")
        self.elbv2_backend.delete_rule(arn)
        return EmptyResult()

    def delete_target_group(self) -> ActionResult:
        arn = self._get_param("TargetGroupArn")
        self.elbv2_backend.delete_target_group(arn)
        return EmptyResult()

    def delete_listener(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        self.elbv2_backend.delete_listener(arn)
        return EmptyResult()

    def modify_rule(self) -> ActionResult:
        rule_arn = self._get_param("RuleArn")
        params = self._get_params()
        conditions = params.get("Conditions", [])
        actions = params.get("Actions", [])
        rule = self.elbv2_backend.modify_rule(
            rule_arn=rule_arn, conditions=conditions, actions=actions
        )
        result = {"Rules": [rule]}
        return ActionResult(result)

    def modify_target_group_attributes(self) -> ActionResult:
        target_group_arn = self._get_param("TargetGroupArn")
        attrs = self._get_param("Attributes", [])
        attributes = {attr["Key"]: attr["Value"] for attr in attrs}
        self.elbv2_backend.modify_target_group_attributes(target_group_arn, attributes)
        result = {"Attributes": attributes}
        return ActionResult(result)

    def register_targets(self) -> ActionResult:
        target_group_arn = self._get_param("TargetGroupArn")
        targets = self._get_param("Targets", [])
        self.elbv2_backend.register_targets(target_group_arn, targets)
        return EmptyResult()

    def deregister_targets(self) -> ActionResult:
        target_group_arn = self._get_param("TargetGroupArn")
        targets = self._get_param("Targets", [])
        self.elbv2_backend.deregister_targets(target_group_arn, targets)
        return EmptyResult()

    def describe_target_health(self) -> ActionResult:
        target_group_arn = self._get_param("TargetGroupArn")
        targets = self._get_param("Targets", [])
        target_health_descriptions = self.elbv2_backend.describe_target_health(
            target_group_arn, targets
        )
        result = {"TargetHealthDescriptions": target_health_descriptions}
        return ActionResult(result)

    def set_rule_priorities(self) -> ActionResult:
        rule_priorities = self._get_param("RulePriorities", [])
        rules = self.elbv2_backend.set_rule_priorities(rule_priorities)
        result = {"Rules": rules}
        return ActionResult(result)

    def add_tags(self) -> ActionResult:
        resource_arns = self._get_param("ResourceArns", [])
        tags = self._get_param("Tags", [])
        self.elbv2_backend.add_tags(resource_arns, tags)  # type: ignore
        return EmptyResult()

    def remove_tags(self) -> ActionResult:
        resource_arns = self._get_param("ResourceArns", [])
        tag_keys = self._get_param("TagKeys", [])
        self.elbv2_backend.remove_tags(resource_arns, tag_keys)
        return EmptyResult()

    def describe_tags(self) -> ActionResult:
        resource_arns = self._get_param("ResourceArns", [])
        resource_tags = self.elbv2_backend.describe_tags(resource_arns)
        result = {
            "TagDescriptions": [
                {"ResourceArn": arn, "Tags": tags}
                for arn, tags in resource_tags.items()
            ]
        }
        return ActionResult(result)

    def describe_account_limits(self) -> ActionResult:
        # Supports paging but not worth implementing yet
        # marker = self._get_param('Marker')
        # page_size = self._get_int_param('PageSize')

        limit_data = {
            "application-load-balancers": 20,
            "target-groups": 3000,
            "targets-per-application-load-balancer": 30,
            "listeners-per-application-load-balancer": 50,
            "rules-per-application-load-balancer": 100,
            "network-load-balancers": 20,
            "targets-per-network-load-balancer": 200,
            "listeners-per-network-load-balancer": 50,
            "certificates-per-application-load-balancer": 25,
        }
        limits = [{"Name": k, "Max": v} for k, v in limit_data.items()]
        result = {"Limits": limits}
        return ActionResult(result)

    def describe_ssl_policies(self) -> ActionResult:
        names = self._get_param("Names", [])
        # Supports paging but not worth implementing yet
        # marker = self._get_param('Marker')
        # page_size = self._get_int_param('PageSize')

        policies = []
        seen = set()

        if names:
            for name in names:
                if name in SSL_POLICIES and name not in seen:
                    policies.append(build_policy(name, SSL_POLICIES[name]))
                    seen.add(name)
        else:
            for name, data in SSL_POLICIES.items():
                policies.append(build_policy(name, data))

        result = {"SslPolicies": policies}
        return ActionResult(result)

    def set_ip_address_type(self) -> ActionResult:
        arn = self._get_param("LoadBalancerArn")
        ip_type = self._get_param("IpAddressType")
        self.elbv2_backend.set_ip_address_type(arn, ip_type)
        result = {"IpAddressType": ip_type}
        return ActionResult(result)

    def set_security_groups(self) -> ActionResult:
        arn = self._get_param("LoadBalancerArn")
        sec_groups = self._get_param("SecurityGroups", [])
        self.elbv2_backend.set_security_groups(arn, sec_groups)
        result = {"SecurityGroups": sec_groups}
        return ActionResult(result)

    def set_subnets(self) -> ActionResult:
        arn = self._get_param("LoadBalancerArn")
        subnets = self._get_param("Subnets", [])
        subnet_mappings = self._get_param("SubnetMappings", [])
        subnet_zone_list = self.elbv2_backend.set_subnets(arn, subnets, subnet_mappings)
        result = {"AvailabilityZones": subnet_zone_list}
        return ActionResult(result)

    def modify_load_balancer_attributes(self) -> ActionResult:
        arn = self._get_param("LoadBalancerArn")
        attrs = self._get_param("Attributes", [])
        attrs = {attr["Key"]: attr["Value"] for attr in attrs}
        all_attrs = self.elbv2_backend.modify_load_balancer_attributes(arn, attrs)
        result = {"Attributes": all_attrs}
        return ActionResult(result)

    def describe_load_balancer_attributes(self) -> ActionResult:
        arn = self._get_param("LoadBalancerArn")
        attrs = self.elbv2_backend.describe_load_balancer_attributes(arn)
        result = {"Attributes": attrs}
        return ActionResult(result)

    def modify_target_group(self) -> ActionResult:
        arn = self._get_param("TargetGroupArn")

        health_check_proto = self._get_param(
            "HealthCheckProtocol"
        )  # 'HTTP' | 'HTTPS' | 'TCP',
        health_check_port = self._get_param("HealthCheckPort")
        health_check_path = self._get_param("HealthCheckPath")
        health_check_interval = self._get_int_param("HealthCheckIntervalSeconds")
        health_check_timeout = self._get_int_param("HealthCheckTimeoutSeconds")
        health_check_enabled = self._get_param("HealthCheckEnabled")
        healthy_threshold_count = self._get_param("HealthyThresholdCount")
        unhealthy_threshold_count = self._get_param("UnhealthyThresholdCount")
        http_codes = self._get_param("Matcher.HttpCode")
        target_group = self.elbv2_backend.modify_target_group(
            arn,
            health_check_proto,
            health_check_port,
            health_check_path,
            health_check_interval,
            health_check_timeout,
            healthy_threshold_count,
            unhealthy_threshold_count,
            http_codes,
            health_check_enabled=health_check_enabled,
        )
        result = {"TargetGroups": [target_group]}
        return ActionResult(result)

    def modify_listener(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        port = self._get_param("Port")
        protocol = self._get_param("Protocol")
        ssl_policy = self._get_param("SslPolicy")
        certificates = self._get_param("Certificates", [])
        default_actions = self._get_param("DefaultActions", [])

        if ssl_policy is not None and ssl_policy not in SSL_POLICIES:
            raise RESTError("SSLPolicyNotFound", f"Policy {ssl_policy} not found")

        listener = self.elbv2_backend.modify_listener(
            arn, port, protocol, ssl_policy, certificates, default_actions
        )

        result = {"Listeners": [listener]}
        return ActionResult(result)

    def add_listener_certificates(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        certificates = self._get_param("Certificates", [])
        certificate_arns = self.elbv2_backend.add_listener_certificates(
            arn, certificates
        )
        result = {"Certificates": [{"CertificateArn": arn} for arn in certificate_arns]}
        return ActionResult(result)

    def describe_listener_certificates(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        certificates = self.elbv2_backend.describe_listener_certificates(arn)
        # The first certificate is the default, others are SNI certificates
        result = {
            "Certificates": [
                {"CertificateArn": cert_arn, "IsDefault": idx == 0}
                for idx, cert_arn in enumerate(certificates)
            ]
        }
        return ActionResult(result)

    def remove_listener_certificates(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        certificates = self._get_param("Certificates", [])
        self.elbv2_backend.remove_listener_certificates(arn, certificates)
        return EmptyResult()

    def describe_listener_attributes(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        attrs = self.elbv2_backend.describe_listener_attributes(arn)
        result = {"Attributes": attrs}
        return ActionResult(result)

    def modify_listener_attributes(self) -> ActionResult:
        arn = self._get_param("ListenerArn")
        attrs = self._get_params()["Attributes"]
        updated_attrs = self.elbv2_backend.modify_listener_attributes(
            listener_arn=arn, attrs=attrs
        )
        result = {"Attributes": updated_attrs}
        return ActionResult(result)

    def describe_capacity_reservation(self) -> ActionResult:
        result = {"CapacityReservationState": [{"State": {"Code": "provisioned"}}]}
        return ActionResult(result)



from moto.core.exceptions import ServiceException

from .constants import MAXIMUM_VISIBILITY_TIMEOUT


class SQSException(ServiceException):
    pass


class ReceiptHandleIsInvalid(SQSException):
    def __init__(self) -> None:
        super().__init__(
            "ReceiptHandleIsInvalid", "The input receipt handle is invalid."
        )


class MessageAttributesInvalid(SQSException):
    code = "InvalidParameterValue"

    def __init__(self, description: str):
        super().__init__(description)


class QueueDoesNotExist(SQSException):
    def __init__(self) -> None:
        super().__init__(
            "AWS.SimpleQueueService.NonExistentQueue",
            "The specified queue does not exist.",
        )


class QueueAlreadyExists(SQSException):
    def __init__(self, message: str):
        super().__init__("QueueAlreadyExists", message)


class EmptyBatchRequest(SQSException):
    def __init__(self, action: str = "Send") -> None:
        super().__init__(
            "AWS.SimpleQueueService.EmptyBatchRequest",
            f"There should be at least one {action}MessageBatchRequestEntry in the request.",
        )


class InvalidBatchEntryId(SQSException):
    def __init__(self) -> None:
        super().__init__(
            "InvalidBatchEntryId",
            "A batch entry id can only contain alphanumeric characters, "
            "hyphens and underscores. It can be at most 80 letters long.",
        )


class BatchRequestTooLong(SQSException):
    def __init__(self, length: int):
        # local import to avoid circular dependencies
        from .models import MAXIMUM_MESSAGE_LENGTH

        super().__init__(
            "BatchRequestTooLong",
            f"Batch requests cannot be longer than {MAXIMUM_MESSAGE_LENGTH} bytes. "
            f"You have sent {length} bytes.",
        )


class BatchEntryIdsNotDistinct(SQSException):
    def __init__(self, entry_id: str):
        super().__init__("BatchEntryIdsNotDistinct", f"Id {entry_id} repeated.")


class TooManyEntriesInBatchRequest(SQSException):
    def __init__(self, number: int):
        super().__init__(
            "TooManyEntriesInBatchRequest",
            f"Maximum number of entries per request are 10. You have sent {number}.",
        )


class InvalidAttributeName(SQSException):
    def __init__(self, attribute_name: str):
        super().__init__("InvalidAttributeName", f"Unknown Attribute {attribute_name}.")


class InvalidAttributeValue(SQSException):
    def __init__(self, attribute_name: str):
        super().__init__(
            "InvalidAttributeValue",
            f"Invalid value for the parameter {attribute_name}.",
        )


class InvalidParameterValue(SQSException):
    def __init__(self, message: str):
        super().__init__("InvalidParameterValue", message)


class MaxVisibilityTimeout(SQSException):
    code = "InvalidParameterValue"
    message = (
        f"Invalid request, maximum visibility timeout is {MAXIMUM_VISIBILITY_TIMEOUT}"
    )


class MissingParameter(SQSException):
    def __init__(self, parameter: str):
        super().__init__(
            "MissingParameter", f"The request must contain the parameter {parameter}."
        )


class OverLimit(SQSException):
    def __init__(self, count: int):
        super().__init__(
            "OverLimit", f"{count} Actions were found, maximum allowed is 7."
        )



from moto.core.responses import ActionResult, EmptyResult
from moto.core.utils import camelcase_to_underscores
from moto.moto_api._internal import mock_random as random

from ..exceptions import InvalidParameterCombination, MissingParameter
from ._base_response import EC2BaseResponse


class Subnets(EC2BaseResponse):
    def create_subnet(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        cidr_block = self._get_param("CidrBlock")
        ipv4_ipam_pool_id = self._get_param("Ipv4IpamPoolId")
        ipv6_cidr_block = self._get_param("Ipv6CidrBlock")
        ipv6_ipam_pool_id = self._get_param("Ipv6IpamPoolId")
        availability_zone = self._get_param("AvailabilityZone")
        availability_zone_id = self._get_param("AvailabilityZoneId")
        tags = self._parse_tag_specification("subnet")
        ipv6_native = self._get_bool_param("Ipv6Native", False)

        if ipv6_native and not ipv6_cidr_block and not ipv6_ipam_pool_id:
            raise MissingParameter(
                "Either 'ipv6CidrBlock' or 'ipv6IpamPoolId' should be provided."
            )

        if ipv6_native and (cidr_block or ipv4_ipam_pool_id):
            raise InvalidParameterCombination(
                "When specifying ipv4 parameters, cidrBlock or ipv4IpamPoolId, you cannot set ipv6Native to true."
            )

        if not availability_zone and not availability_zone_id:
            availability_zone = random.choice(
                self.ec2_backend.describe_availability_zones()
            ).name
        subnet = self.ec2_backend.create_subnet(
            vpc_id,
            cidr_block,
            ipv6_cidr_block,
            availability_zone,
            availability_zone_id,
            ipv6_native=ipv6_native,
            tags=tags,
        )
        result = {"Subnet": subnet}
        return ActionResult(result)

    def create_default_subnet(self) -> ActionResult:
        availability_zone = self._get_param("AvailabilityZone")

        subnet = self.ec2_backend.create_default_subnet(availability_zone)
        result = {"Subnet": subnet}
        return ActionResult(result)

    def delete_subnet(self) -> ActionResult:
        subnet_id = self._get_param("SubnetId")
        self.ec2_backend.delete_subnet(subnet_id)
        return EmptyResult()

    def describe_subnets(self) -> ActionResult:
        self.error_on_dryrun()
        subnet_ids = self._get_param("SubnetIds")
        filters = self._filters_from_querystring()
        subnets = self.ec2_backend.describe_subnets(subnet_ids, filters)
        result = {"Subnets": subnets}
        return ActionResult(result)

    def modify_subnet_attribute(self) -> ActionResult:
        subnet_id = self._get_param("SubnetId")

        for attribute in ("MapPublicIpOnLaunch", "AssignIpv6AddressOnCreation"):
            if self._get_param(f"{attribute}.Value") is not None:
                attr_name = camelcase_to_underscores(attribute)
                attr_value = self._get_param(f"{attribute}.Value")
                self.ec2_backend.modify_subnet_attribute(
                    subnet_id, attr_name, attr_value
                )
        return EmptyResult()

    def associate_subnet_cidr_block(self) -> ActionResult:
        ipv6_cidr_block = self._get_param("Ipv6CidrBlock")
        subnet_id = self._get_param("SubnetId")
        association = self.ec2_backend.associate_subnet_cidr_block(
            subnet_id, ipv6_cidr_block
        )
        result = {"Ipv6CidrBlockAssociation": association, "SubnetId": subnet_id}
        return ActionResult(result)

    def disassociate_subnet_cidr_block(self) -> ActionResult:
        association_id = self._get_param("AssociationId")
        subnet_id, association = self.ec2_backend.disassociate_subnet_cidr_block(
            association_id
        )
        result = {"Ipv6CidrBlockAssociation": association, "SubnetId": subnet_id}
        return ActionResult(result)

    def create_subnet_cidr_reservation(self) -> ActionResult:
        self.error_on_dryrun()
        subnet_id = self._get_param("SubnetId")
        reservation_type = self._get_param("ReservationType")
        cidr = self._get_param("Cidr")
        tag_specifications = self._get_param("TagSpecifications", [])

        reservation = self.ec2_backend.create_subnet_cidr_reservation(
            subnet_id, reservation_type, cidr, tag_specifications
        )

        return ActionResult({"SubnetCidrReservation": reservation})

    def get_subnet_cidr_reservations(self) -> ActionResult:
        self.error_on_dryrun()

        subnet_id = self._get_param("SubnetId")
        filters = self._filters_from_querystring()
        cidr_reservations = self.ec2_backend.get_subnet_cidr_reservations(
            subnet_id, filters
        )
        return ActionResult(cidr_reservations)

    def delete_subnet_cidr_reservation(self) -> ActionResult:
        self.error_on_dryrun()

        reservation_id = self._get_param("SubnetCidrReservationId")

        deleted_reservation = self.ec2_backend.delete_subnet_cidr_reservation(
            reservation_id
        )

        return ActionResult({"DeletedSubnetCidrReservation": deleted_reservation})



from moto.core.responses import ActionResult
from moto.ec2.utils import add_tag_specification

from ._base_response import EC2BaseResponse


class CarrierGateway(EC2BaseResponse):
    def create_carrier_gateway(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        tag_param = self._get_param("TagSpecifications", [])
        tags = add_tag_specification(tag_param)

        carrier_gateway = self.ec2_backend.create_carrier_gateway(
            vpc_id=vpc_id, tags=tags
        )
        result = {"CarrierGateway": carrier_gateway}
        return ActionResult(result)

    def delete_carrier_gateway(self) -> ActionResult:
        carrier_gateway_id = self._get_param("CarrierGatewayId")

        carrier_gateway = self.ec2_backend.delete_carrier_gateway(carrier_gateway_id)
        result = {"CarrierGateway": carrier_gateway}
        return ActionResult(result)

    def describe_carrier_gateways(self) -> ActionResult:
        carrier_gateway_ids = self._get_param("CarrierGatewayIds")
        filters = self._filters_from_querystring()

        carrier_gateways = self.ec2_backend.describe_carrier_gateways(
            carrier_gateway_ids, filters
        )
        result = {"CarrierGateways": carrier_gateways}
        return ActionResult(result)



from .responses import MacieResponse

url_bases = [
    r"https?://macie2\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/invitations$": MacieResponse.dispatch,
    "{0}/invitations/decline$": MacieResponse.dispatch,
    "{0}/invitations/accept$": MacieResponse.dispatch,
    "{0}/members$": MacieResponse.dispatch,
    "{0}/members/[^/]+$": MacieResponse.dispatch,
    "{0}/members/(?P<id>[^/]+)$": MacieResponse.dispatch,
    "{0}/administrator$": MacieResponse.dispatch,
    "{0}/macie$": MacieResponse.dispatch,
    "{0}/session$": MacieResponse.dispatch,
    "{0}/members/disassociate/(?P<id>[^/]+)$": MacieResponse.dispatch,
    "{0}/admin$": MacieResponse.dispatch,
}



from moto.core.exceptions import ServiceException


class ValidationError(ServiceException):
    code = "ValidationException"

    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class VectorBucketNotFound(ServiceException):
    code = "NotFoundException"

    def __init__(self) -> None:
        super().__init__("The specified vector bucket could not be found")


class IndexNotFound(ServiceException):
    code = "NotFoundException"

    def __init__(self) -> None:
        super().__init__("The specified index could not be found")


class VectorBucketInvalidLength(ServiceException):
    code = "ValidationException"

    def __init__(self, length: int):
        super().__init__(
            f"1 validation error detected. Value with length {length} at '/vectorBucketName' failed to satisfy constraint: Member must have length between 3 and 63, inclusive"
        )


class VectorBucketInvalidChars(ServiceException):
    code = "ValidationException"

    def __init__(self) -> None:
        super().__init__("Invalid vector bucket name")


class VectorBucketAlreadyExists(ServiceException):
    code = "ConflictException"

    def __init__(self) -> None:
        super().__init__("A vector bucket with the specified name already exists")


class VectorBucketNotEmpty(ServiceException):
    code = "ConflictException"

    def __init__(self) -> None:
        super().__init__("The specified vector bucket is not empty")


class VectorBucketPolicyNotFound(ServiceException):
    code = "NotFoundException"

    def __init__(self) -> None:
        super().__init__("The specified vector bucket policy could not be found")


class VectorWrongDimension(ServiceException):
    code = "ValidationException"

    def __init__(self, key: str, actual: int, provided: int):
        super().__init__(
            f"Invalid record for key '{key}': vector must have length {actual}, but has length {provided}"
        )



"""Handles incoming emrserverless requests, invokes methods, returns responses."""

import json

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse

from .models import EMRServerlessBackend, emrserverless_backends

DEFAULT_MAX_RESULTS = 100
DEFAULT_NEXT_TOKEN = ""

"""
These are the available methods:
    can_paginate()
    cancel_job_run() -> DONE
    close()
    create_application() -> DONE
    delete_application() -> DONE
    get_application() -> DONE
    get_job_run() -> DONE
    get_paginator()
    get_waiter()
    list_applications() -> DONE
    list_job_runs() -> DONE
    list_tags_for_resource()
    start_application() -> DONE
    start_job_run() -> DONE
    stop_application() -> DONE
    tag_resource()
    untag_resource()
    update_application()
"""


class EMRServerlessResponse(BaseResponse):
    """Handler for EMRServerless requests and responses."""

    def __init__(self) -> None:
        super().__init__("emr-serverless")

    @property
    def emrserverless_backend(self) -> EMRServerlessBackend:
        """Return backend instance specific for this region."""
        return emrserverless_backends[self.current_account][self.region]

    def create_application(self) -> TYPE_RESPONSE:
        name = self._get_param("name")
        release_label = self._get_param("releaseLabel")
        application_type = self._get_param("type")
        client_token = self._get_param("clientToken")
        initial_capacity = self._get_param("initialCapacity")
        maximum_capacity = self._get_param("maximumCapacity")
        tags = self._get_param("tags")
        auto_start_configuration = self._get_param("autoStartConfiguration")
        auto_stop_configuration = self._get_param("autoStopConfiguration")
        network_configuration = self._get_param("networkConfiguration")

        application = self.emrserverless_backend.create_application(
            name=name,
            release_label=release_label,
            application_type=application_type,
            client_token=client_token,
            initial_capacity=initial_capacity,
            maximum_capacity=maximum_capacity,
            tags=tags,
            auto_start_configuration=auto_start_configuration,
            auto_stop_configuration=auto_stop_configuration,
            network_configuration=network_configuration,
        )
        return 200, {}, json.dumps(dict(application))

    def delete_application(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")

        self.emrserverless_backend.delete_application(application_id=application_id)
        return 200, {}, ""

    def get_application(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")

        application = self.emrserverless_backend.get_application(
            application_id=application_id
        )
        response = {"application": application}
        return 200, {}, json.dumps(response)

    def list_applications(self) -> TYPE_RESPONSE:
        states = self.querystring.get("states", [])
        max_results = self._get_int_param("maxResults", DEFAULT_MAX_RESULTS)
        next_token = self._get_param("nextToken", DEFAULT_NEXT_TOKEN)

        applications, next_token = self.emrserverless_backend.list_applications(
            next_token=next_token,
            max_results=max_results,
            states=states,
        )
        response = {"applications": applications, "nextToken": next_token}
        return 200, {}, json.dumps(response)

    def start_application(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")

        self.emrserverless_backend.start_application(application_id=application_id)
        return 200, {}, ""

    def stop_application(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")

        self.emrserverless_backend.stop_application(application_id=application_id)
        return 200, {}, ""

    def update_application(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")
        initial_capacity = self._get_param("initialCapacity")
        maximum_capacity = self._get_param("maximumCapacity")
        auto_start_configuration = self._get_param("autoStartConfiguration")
        auto_stop_configuration = self._get_param("autoStopConfiguration")
        network_configuration = self._get_param("networkConfiguration")

        application = self.emrserverless_backend.update_application(
            application_id=application_id,
            initial_capacity=initial_capacity,
            maximum_capacity=maximum_capacity,
            auto_start_configuration=auto_start_configuration,
            auto_stop_configuration=auto_stop_configuration,
            network_configuration=network_configuration,
        )
        response = {"application": application}
        return 200, {}, json.dumps(response)

    def start_job_run(self) -> TYPE_RESPONSE:
        # params = self._get_params()
        application_id = self._get_param(
            "applicationId"
        )  # params["application_id"] #.get()
        client_token = self._get_param("clientToken")
        execution_role_arn = self._get_param("executionRoleArn")
        job_driver = self._get_param("jobDriver")
        configuration_overrides = self._get_param("configurationOverrides")
        tags = self._get_param("tags")
        execution_timeout_minutes = self._get_int_param("executionTimeoutMinutes")
        name = self._get_param("name")
        job_run = self.emrserverless_backend.start_job_run(
            application_id=application_id,
            client_token=client_token,
            execution_role_arn=execution_role_arn,
            job_driver=job_driver,
            configuration_overrides=configuration_overrides,
            tags=tags,
            execution_timeout_minutes=execution_timeout_minutes,
            name=name,
        )
        return (
            200,
            {},
            json.dumps(
                {
                    "applicationId": job_run.application_id,
                    "jobRunId": job_run.id,
                    "arn": job_run.arn,
                }
            ),
        )

    def get_job_run(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")
        job_run_id = self._get_param("jobRunId")
        job_run = self.emrserverless_backend.get_job_run(
            application_id=application_id, job_run_id=job_run_id
        )
        # TODO: adjust response
        return 200, {}, json.dumps({"jobRun": job_run.to_dict("get")})

    def cancel_job_run(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")
        job_run_id = self._get_param("jobRunId")
        application_id, job_run_id = self.emrserverless_backend.cancel_job_run(
            application_id=application_id,
            job_run_id=job_run_id,
        )

        return (
            200,
            {},
            json.dumps({"applicationId": application_id, "jobRunId": job_run_id}),
        )

    def list_job_runs(self) -> TYPE_RESPONSE:
        application_id = self._get_param("applicationId")
        next_token = self._get_param("nextToken", DEFAULT_NEXT_TOKEN)
        max_results = self._get_int_param("maxResults", DEFAULT_MAX_RESULTS)
        created_at_after = self._get_param("createdAtAfter")
        created_at_before = self._get_param("createdAtBefore")
        states = self.querystring.get("states", [])
        job_runs, next_token = self.emrserverless_backend.list_job_runs(
            application_id=application_id,
            next_token=next_token,
            max_results=max_results,
            created_at_after=created_at_after,
            created_at_before=created_at_before,
            states=states,
        )
        return 200, {}, json.dumps({"jobRuns": job_runs, "nextToken": next_token})



from moto.core.responses import ActionResult
from moto.core.utils import utcnow
from moto.ec2.utils import Base64EncodedString

from ._base_response import EC2BaseResponse

SAMPLE_CONSOLE_OUTPUT = """Linux version 2.6.16-xenU (builder@patchbat.amazonsa) (gcc version 4.0.1 20050727 (Red Hat 4.0.1-5)) #1 SMP Thu Oct 26 08:41:26 SAST 2006
BIOS-provided physical RAM map:
Xen: 0000000000000000 - 000000006a400000 (usable)
980MB HIGHMEM available.
727MB LOWMEM available.
NX (Execute Disable) protection: active
IRQ lockup detection disabled
Built 1 zonelists
Kernel command line: root=/dev/sda1 ro 4
Enabling fast FPU save and restore... done.
"""


class General(EC2BaseResponse):
    def get_console_output(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        instance = self.ec2_backend.get_instance(instance_id)
        result = {
            "InstanceId": instance.id,
            "Timestamp": utcnow(),
            "Output": Base64EncodedString.from_raw_string(SAMPLE_CONSOLE_OUTPUT),
        }
        return ActionResult(result)



import uuid
from collections import defaultdict
from datetime import datetime, timezone
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.moto_api._internal import mock_random
from moto.s3 import s3_backends
from moto.s3.exceptions import (
    InvalidPublicAccessBlockConfiguration,
    WrongPublicAccessBlockAccountIdError,
)
from moto.s3.models import PublicAccessBlock, S3Backend
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import PARTITION_NAMES, get_partition

from .exceptions import (
    AccessPointNotFound,
    AccessPointPolicyNotFound,
    InvalidRequestException,
    MultiRegionAccessPointNotFound,
    MultiRegionAccessPointOperationNotFound,
    MultiRegionAccessPointPolicyNotFound,
    NoSuchPublicAccessBlockConfiguration,
    StorageLensConfigurationNotFound,
)

PAGINATION_MODEL = {
    "list_storage_lens_configurations": {
        "input_token": "next_token",
        "limit_default": 100,
        "unique_attribute": "id",
    },
    "list_access_points": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 1000,
        "unique_attribute": "name",
    },
    "list_multi_region_access_points": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "name",
    },
}


class AccessPoint(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        name: str,
        bucket: str,
        vpc_configuration: dict[str, Any],
        public_access_block_configuration: dict[str, Any],
    ):
        self.name = name
        self.alias = f"{name}-{mock_random.get_random_hex(34)}-s3alias"
        self.bucket = bucket
        self.created = datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.%f")
        self.arn = f"arn:{get_partition(region_name)}:s3:us-east-1:{account_id}:accesspoint/{name}"
        self.policy: Optional[str] = None
        self.network_origin = "VPC" if vpc_configuration else "Internet"
        self.vpc_id = (vpc_configuration or {}).get("VpcId")
        pubc = public_access_block_configuration or {}
        self.pubc = {
            "BlockPublicAcls": pubc.get("BlockPublicAcls", "true"),
            "IgnorePublicAcls": pubc.get("IgnorePublicAcls", "true"),
            "BlockPublicPolicy": pubc.get("BlockPublicPolicy", "true"),
            "RestrictPublicBuckets": pubc.get("RestrictPublicBuckets", "true"),
        }

    def delete_policy(self) -> None:
        self.policy = None

    def set_policy(self, policy: str) -> None:
        self.policy = policy

    def has_policy(self) -> bool:
        return self.policy is not None


class MultiRegionAccessPoint(BaseModel):
    def __init__(
        self,
        name: str,
        public_access_block: dict[str, Any],
        regions: list[dict[str, str]],
    ):
        self.name = name
        self.alias = f"{name}-{mock_random.get_random_hex(10)}.mrap"
        self.created_at = datetime.now(timezone.utc)
        self.public_access_block = public_access_block or {
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        }
        self.regions = regions
        self.status = "READY"
        self.policy: Optional[str] = None

    def set_policy(self, policy: str) -> None:
        self.policy = policy

    def has_policy(self) -> bool:
        return self.policy is not None

    def to_dict(self) -> dict[str, Any]:
        return {
            "Name": self.name or "",
            "Alias": self.alias,
            "CreatedAt": self.created_at.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "PublicAccessBlock": self.public_access_block,
            "Status": self.status,
            "Regions": self.regions,
        }


class MultiRegionAccessPointOperation(BaseModel):
    def __init__(
        self,
        account_id: str,
        operation: str,
        region_name: str,
        name: Optional[str] = None,
        details: Optional[dict[str, Any]] = None,
    ):
        self.request_token_arn = f"arn:aws:s3:{region_name}:{account_id}:async-request/mrap/{operation.lower()}/{uuid.uuid4().hex[:24]}"
        self.request_status = "SUCCEEDED"
        self.name = name
        self.operation = operation
        self.created_at = datetime.now(timezone.utc)
        self.details = details or {}

    def to_dict(self) -> dict[str, Any]:
        response: dict[str, Any] = {
            "RequestTokenARN": self.request_token_arn,
            "RequestStatus": self.request_status,
            "CreationTime": self.created_at.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "Operation": self.operation,
        }

        if self.operation == "CreateMultiRegionAccessPoint" and self.details:
            response["RequestParameters"] = {
                "CreateMultiRegionAccessPointRequest": {
                    "Name": self.name or "",
                    "Regions": self.details.get("Regions", []),
                    "PublicAccessBlock": self.details.get("PublicAccessBlock", {}),
                }
            }
            regions_response = [
                {"Name": r.get("Region", ""), "RequestStatus": self.request_status}
                for r in self.details.get("Regions", [])
            ]
            response["ResponseDetails"] = {
                "MultiRegionAccessPointDetails": {
                    "Regions": regions_response,
                }
            }
        elif self.operation == "DeleteMultiRegionAccessPoint":
            response["RequestParameters"] = {
                "DeleteMultiRegionAccessPointRequest": {
                    "Name": self.name or "",
                }
            }
            response["ResponseDetails"] = {
                "MultiRegionAccessPointDetails": {
                    "Regions": [],
                }
            }
        elif self.operation == "PutMultiRegionAccessPointPolicy":
            response["RequestParameters"] = {
                "PutMultiRegionAccessPointPolicyRequest": {
                    "Name": self.name or "",
                    "Policy": self.details.get("Policy", ""),
                }
            }
            response["ResponseDetails"] = {}

        return response


class StorageLensConfiguration(BaseModel):
    def __init__(
        self,
        account_id: str,
        config_id: str,
        storage_lens_configuration: dict[str, Any],
        tags: Optional[dict[str, str]] = None,
    ):
        self.account_id = account_id
        self.config_id = config_id
        self.config = storage_lens_configuration
        self.tags = tags or {}
        self.arn = f"arn:{get_partition('us-east-1')}:s3:us-east-1:{account_id}:storage-lens/{config_id}"


class S3ControlBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.public_access_block: Optional[PublicAccessBlock] = None
        self.access_points: dict[str, dict[str, AccessPoint]] = defaultdict(dict)
        self.multi_region_access_points: dict[
            str, dict[str, MultiRegionAccessPoint]
        ] = defaultdict(dict)
        self.mrap_operations: dict[str, dict[str, MultiRegionAccessPointOperation]] = (
            defaultdict(dict)
        )
        self.storage_lens_configs: dict[str, StorageLensConfiguration] = {}
        self.tagger = TaggingService()

    def get_public_access_block(self, account_id: str) -> PublicAccessBlock:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()

        if not self.public_access_block:
            raise NoSuchPublicAccessBlockConfiguration()

        return self.public_access_block

    def delete_public_access_block(self, account_id: str) -> None:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()

        self.public_access_block = None

    def put_public_access_block(
        self, account_id: str, pub_block_config: dict[str, Any]
    ) -> None:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()

        if not pub_block_config:
            raise InvalidPublicAccessBlockConfiguration()

        self.public_access_block = PublicAccessBlock(
            pub_block_config.get("BlockPublicAcls"),
            pub_block_config.get("IgnorePublicAcls"),
            pub_block_config.get("BlockPublicPolicy"),
            pub_block_config.get("RestrictPublicBuckets"),
        )

    def create_access_point(
        self,
        account_id: str,
        name: str,
        bucket: str,
        vpc_configuration: dict[str, Any],
        public_access_block_configuration: dict[str, Any],
    ) -> AccessPoint:
        access_point = AccessPoint(
            account_id,
            region_name=self.region_name,
            name=name,
            bucket=bucket,
            vpc_configuration=vpc_configuration,
            public_access_block_configuration=public_access_block_configuration,
        )
        self.access_points[account_id][name] = access_point
        return access_point

    def delete_access_point(self, account_id: str, name: str) -> None:
        self.access_points[account_id].pop(name, None)

    def get_access_point(self, account_id: str, name: str) -> AccessPoint:
        if name not in self.access_points[account_id]:
            raise AccessPointNotFound(name)
        return self.access_points[account_id][name]

    def put_access_point_policy(self, account_id: str, name: str, policy: str) -> None:
        access_point = self.get_access_point(account_id, name)
        access_point.set_policy(policy)

    def get_access_point_policy(self, account_id: str, name: str) -> str:
        access_point = self.get_access_point(account_id, name)
        if access_point.has_policy():
            return access_point.policy  # type: ignore[return-value]
        raise AccessPointPolicyNotFound(name)

    def delete_access_point_policy(self, account_id: str, name: str) -> None:
        access_point = self.get_access_point(account_id, name)
        access_point.delete_policy()

    def get_access_point_policy_status(self, account_id: str, name: str) -> bool:
        self.get_access_point_policy(account_id, name)
        return True

    def create_multi_region_access_point(
        self,
        account_id: str,
        name: str,
        public_access_block: dict[str, Any],
        regions: list[dict[str, str]],
        region_name: str,
    ) -> MultiRegionAccessPointOperation:
        if name in self.multi_region_access_points[account_id]:
            raise InvalidRequestException(
                f"Multi-Region Access Point {name} already exists"
            )

        processed_regions: list[dict[str, str]] = []
        for region_item in regions:
            bucket_name = region_item.get("Bucket", "")
            found_region = "us-east-1"

            found = False
            for account_id_key in s3_backends:
                if found:
                    break
                for region_key, backend in s3_backends[account_id_key].items():
                    if region_key == "aws":
                        continue

                    if bucket_name in backend.buckets:
                        found_region = region_key
                        found = True
                        break

            processed_regions.append({"Bucket": bucket_name, "Region": found_region})

        mrap = MultiRegionAccessPoint(
            name=name,
            public_access_block=public_access_block,
            regions=processed_regions,
        )
        self.multi_region_access_points[account_id][name] = mrap

        operation = MultiRegionAccessPointOperation(
            account_id=account_id,
            operation="CreateMultiRegionAccessPoint",
            region_name=region_name,
            name=name,
            details={
                "Regions": processed_regions,
                "PublicAccessBlock": public_access_block,
            },
        )
        self.mrap_operations[account_id][operation.request_token_arn] = operation

        return operation

    def delete_multi_region_access_point(
        self,
        account_id: str,
        name: str,
        region_name: str,
    ) -> MultiRegionAccessPointOperation:
        if name not in self.multi_region_access_points[account_id]:
            raise MultiRegionAccessPointNotFound(name)

        del self.multi_region_access_points[account_id][name]

        operation = MultiRegionAccessPointOperation(
            account_id=account_id,
            operation="DeleteMultiRegionAccessPoint",
            region_name=region_name,
            name=name,
        )
        self.mrap_operations[account_id][operation.request_token_arn] = operation

        return operation

    def describe_multi_region_access_point_operation(
        self,
        account_id: str,
        request_token_arn: str,
    ) -> MultiRegionAccessPointOperation:
        if request_token_arn not in self.mrap_operations[account_id]:
            raise MultiRegionAccessPointOperationNotFound(request_token_arn)

        return self.mrap_operations[account_id][request_token_arn]

    def get_multi_region_access_point(
        self,
        account_id: str,
        name: str,
    ) -> MultiRegionAccessPoint:
        if name not in self.multi_region_access_points[account_id]:
            raise MultiRegionAccessPointNotFound(name)

        return self.multi_region_access_points[account_id][name]

    def get_multi_region_access_point_policy(
        self,
        account_id: str,
        name: str,
    ) -> str:
        mrap = self.get_multi_region_access_point(account_id, name)
        if not mrap.has_policy():
            raise MultiRegionAccessPointPolicyNotFound(name)
        return mrap.policy  # type: ignore[return-value]

    def _is_policy_public(
        self, policy: str, public_access_block: dict[str, Any]
    ) -> bool:
        block_public = public_access_block.get("BlockPublicPolicy")
        if block_public is True or block_public == "true":
            return False

        policy_no_spaces = policy.replace(" ", "")
        if '"Principal":"*"' in policy_no_spaces:
            return True
        if '"Principal":{"AWS":"*"}' in policy_no_spaces:
            return True

        return False

    def get_multi_region_access_point_policy_status(
        self,
        account_id: str,
        name: str,
    ) -> dict[str, Any]:
        mrap = self.get_multi_region_access_point(account_id, name)
        if not mrap.has_policy():
            raise MultiRegionAccessPointPolicyNotFound(name)
        is_public = self._is_policy_public(mrap.policy, mrap.public_access_block)  # type: ignore[arg-type]
        return {"IsPublic": is_public}

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_multi_region_access_points(
        self,
        account_id: str,
        max_results: Optional[int] = None,
        next_token: Optional[str] = None,
    ) -> list[MultiRegionAccessPoint]:
        return list(self.multi_region_access_points[account_id].values())

    def put_multi_region_access_point_policy(
        self,
        account_id: str,
        name: str,
        policy: str,
        region_name: str,
    ) -> MultiRegionAccessPointOperation:
        mrap = self.get_multi_region_access_point(account_id, name)
        mrap.set_policy(policy)

        operation = MultiRegionAccessPointOperation(
            account_id=account_id,
            operation="PutMultiRegionAccessPointPolicy",
            region_name=region_name,
            name=name,
            details={"Policy": policy},
        )
        self.mrap_operations[account_id][operation.request_token_arn] = operation

        return operation

    def put_storage_lens_configuration(
        self,
        config_id: str,
        account_id: str,
        storage_lens_configuration: dict[str, Any],
        tags: Optional[dict[str, str]] = None,
    ) -> None:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()

        storage_lens = StorageLensConfiguration(
            account_id=account_id,
            config_id=config_id,
            storage_lens_configuration=storage_lens_configuration,
            tags=tags,
        )
        self.storage_lens_configs[config_id] = storage_lens

    def get_storage_lens_configuration(
        self, config_id: str, account_id: str
    ) -> StorageLensConfiguration:
        if config_id not in self.storage_lens_configs:
            raise StorageLensConfigurationNotFound(config_id)
        storage_lens_configuration = self.storage_lens_configs[config_id]
        return storage_lens_configuration

    def delete_storage_lens_configuration(
        self, config_id: str, account_id: str
    ) -> None:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()
        if config_id not in self.storage_lens_configs:
            raise StorageLensConfigurationNotFound(config_id)
        del self.storage_lens_configs[config_id]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_storage_lens_configurations(
        self, account_id: str
    ) -> list[StorageLensConfiguration]:
        storage_lens_configuration_list = list(self.storage_lens_configs.values())
        return storage_lens_configuration_list

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_access_points(
        self,
        account_id: str,
        bucket: Optional[str] = None,
        max_results: Optional[int] = None,
        next_token: Optional[str] = None,
    ) -> list[AccessPoint]:
        account_access_points = self.access_points.get(account_id, {})
        all_access_points = list(account_access_points.values())

        if bucket:
            return [ap for ap in all_access_points if ap.bucket == bucket]
        return all_access_points

    def put_storage_lens_configuration_tagging(
        self, config_id: str, account_id: str, tags: dict[str, str]
    ) -> None:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()

        if config_id not in self.storage_lens_configs:
            raise AccessPointNotFound(config_id)

        self.storage_lens_configs[config_id].tags = tags

    def get_storage_lens_configuration_tagging(
        self, config_id: str, account_id: str
    ) -> dict[str, str]:
        if account_id != self.account_id:
            raise WrongPublicAccessBlockAccountIdError()
        if config_id not in self.storage_lens_configs:
            raise AccessPointNotFound(config_id)

        return self.storage_lens_configs[config_id].tags

    def list_tags_for_resource(self, resource_arn: str) -> list[dict[str, str]]:
        backend: S3Backend = s3_backends[self.account_id][self.partition]
        return backend.tagger.list_tags_for_resource(resource_arn)["Tags"]

    def tag_resource(self, resource_arn: str, tags: list[dict[str, str]]) -> None:
        backend: S3Backend = s3_backends[self.account_id][self.partition]
        backend.tagger.tag_resource(resource_arn, tags=tags)

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        backend: S3Backend = s3_backends[self.account_id][self.partition]
        backend.tagger.untag_resource_using_names(resource_arn, tag_names=tag_keys)


s3control_backends = BackendDict(
    S3ControlBackend,
    "s3control",
    use_boto3_regions=False,
    additional_regions=PARTITION_NAMES,
)



"""Exceptions raised by the connectcampaigns service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    """When a resource is not found."""

    code = 404

    def __init__(self, message: str) -> None:
        super().__init__("ResourceNotFoundException", message)


class ValidationException(JsonRESTError):
    """When validation fails on input parameters."""

    code = 400

    def __init__(self, message: str) -> None:
        super().__init__("ValidationException", message)



from moto.core.responses import ActionResult, BaseResponse

from .exceptions import STSValidationError
from .models import STSBackend, sts_backends

MAX_FEDERATION_TOKEN_POLICY_LENGTH = 2048
MAX_ROLE_NAME_LENGTH = 64


class TokenResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="sts")
        self.automated_parameter_parsing = True

    @property
    def backend(self) -> STSBackend:
        return sts_backends[self.current_account][self.partition]

    def _determine_resource(self) -> str:
        if "AssumeRole" in self.querystring.get("Action", []):
            return self.querystring.get("RoleArn")[0]  # type: ignore[index]
        return "*"

    def get_session_token(self) -> ActionResult:
        duration = self._get_int_param("DurationSeconds", 43200)
        token = self.backend.get_session_token(duration=duration)
        result = {
            "Credentials": {
                "AccessKeyId": "AKIAIOSFODNN7EXAMPLE",
                "SecretAccessKey": "wJalrXUtnFEMI/K7MDENG/bPxRfiCYzEXAMPLEKEY",
                "SessionToken": "AQoEXAMPLEH4aoAH0gNCAPyJxz4BlCFFxWNE1OPTgk5TthT+FvwqnKwRcOIfrRh3c/LTo6UDdyJwOOvEVPvLXCrrrUtdnniCEXAMPLE/IvU1dYUg2RVAJBanLiHb4IgRmpRV3zrkuWJOgQs8IZZaIv2BXIa2R4OlgkBN9bkUDNCJiBeb/AXlzBBko7b15fjrBs2+cTQtpZ3CYWFXG8C5zqx37wnOE49mRl/+OtkIKGO7fAE",
                "Expiration": token.expiration,
            }
        }
        return ActionResult(result)

    def get_federation_token(self) -> ActionResult:
        duration = self._get_int_param("DurationSeconds", 43200)
        policy = self._get_param("Policy")
        if policy is not None and len(policy) > MAX_FEDERATION_TOKEN_POLICY_LENGTH:
            raise STSValidationError(
                "1 validation error detected: Value "
                '\'{"Version": "2012-10-17", "Statement": [...]}\' '
                "at 'policy' failed to satisfy constraint: Member must have length less than or "
                f" equal to {MAX_FEDERATION_TOKEN_POLICY_LENGTH}"
            )
        name = self._get_param("Name")
        token = self.backend.get_federation_token(duration=duration, name=name)
        result = {
            "Credentials": {
                "AccessKeyId": "AKIAIOSFODNN7EXAMPLE",
                "SecretAccessKey": "wJalrXUtnFEMI/K7MDENG/bPxRfiCYzEXAMPLEKEY",
                "SessionToken": "AQoDYXdzEPT//////////wEXAMPLEtc764bNrC9SAPBSM22wDOk4x4HIZ8j4FZTwdQWLWsKWHGBuFqwAeMicRXmxfpSPfIeoIYRqTflfKD8YUuwthAx7mSEI/qkPpKPi/kMcGdQrmGdeehM4IC1NtBmUpp2wUE8phUZampKsburEDy0KPkyQDYwT7WZ0wq5VSXDvp75YU9HFvlRd8Tx6q6fE8YQcHNVXAkiY9q6d+xo0rKwT38xVqr7ZD0u0iPPkUL64lIZbqBAz+scqKmlzm8FDrypNC9Yjc8fPOLn9FX9KSYvKTr4rvx3iSIlTJabIQwj2ICCR/oLxBA==",
                "Expiration": token.expiration,
            },
            "FederatedUser": {
                "FederatedUserId": f"{self.current_account}:{token.name}",
                "Arn": f"arn:{self.partition}:sts::{self.current_account}:federated-user/{token.name}",
            },
            "PackedPolicySize": 6,
        }
        return ActionResult(result)

    def assume_role(self) -> ActionResult:
        role_session_name = self._get_param("RoleSessionName")
        role_arn = self._get_param("RoleArn")
        policy = self._get_param("Policy")
        duration = self._get_int_param("DurationSeconds", 3600)
        external_id = self._get_param("ExternalId")
        if role_session_name is not None and len(role_session_name) > 64:
            raise STSValidationError(
                "1 validation error detected: Value "
                f"'{role_session_name}' "
                "at 'roleSessionName' failed to satisfy constraint: Member must have length less than or "
                f" equal to {MAX_ROLE_NAME_LENGTH}"
            )
        role = self.backend.assume_role(
            region_name=self.region,
            role_session_name=role_session_name,
            role_arn=role_arn,
            policy=policy,
            duration=duration,
            external_id=external_id,
        )
        result = {
            "Credentials": role,
            "AssumedRoleUser": {
                "AssumedRoleId": role.user_id,
                "Arn": role.arn,
            },
            "PackedPolicySize": 6,
        }
        return ActionResult(result)

    def assume_role_with_web_identity(self) -> ActionResult:
        role_session_name = self._get_param("RoleSessionName")
        role_arn = self._get_param("RoleArn")
        policy = self._get_param("Policy")
        duration = self._get_int_param("DurationSeconds", 3600)
        external_id = self._get_param("ExternalId")
        role = self.backend.assume_role_with_web_identity(
            region_name=self.region,
            role_session_name=role_session_name,
            role_arn=role_arn,
            policy=policy,
            duration=duration,
            external_id=external_id,
        )
        result = {
            "Credentials": role,
            "AssumedRoleUser": {
                "AssumedRoleId": f"ARO123EXAMPLE123:{role.session_name}",
                "Arn": role.arn,
            },
            "PackedPolicySize": 6,
        }
        return ActionResult(result)

    def assume_role_with_saml(self) -> ActionResult:
        role_arn = self._get_param("RoleArn")
        principal_arn = self._get_param("PrincipalArn")
        saml_assertion = self._get_param("SAMLAssertion")
        role = self.backend.assume_role_with_saml(
            role_arn=role_arn,
            principal_arn=principal_arn,
            saml_assertion=saml_assertion,
        )
        result = {
            "Credentials": role,
            "AssumedRoleUser": {
                "AssumedRoleId": role.user_id,
                "Arn": role.arn,
            },
            "PackedPolicySize": 123,
            "Subject": role.user_id,
            "SubjectType": "persistent",
            "Issuer": "http://localhost:3000/",
            "Audience": "https://signin.aws.amazon.com/saml",
            "NameQualifier": "B64EncodedStringOfHashOfIssuerAccountIdAndUserId=",
        }
        return ActionResult(result)

    def get_caller_identity(self) -> ActionResult:
        access_key_id = self.get_access_key()
        user_id, arn, account_id = self.backend.get_caller_identity(
            access_key_id, self.region
        )
        result = {
            "UserId": user_id,
            "Account": account_id,
            "Arn": arn,
        }
        return ActionResult(result)



from moto.core.responses import ActionResult

from ._base_response import EC2BaseResponse


class KeyPairs(EC2BaseResponse):
    def create_key_pair(self) -> ActionResult:
        key_name = self._get_param("KeyName")
        key_type = self._get_param("KeyType")
        tags = self._parse_tag_specification("key-pair").get("key-pair", {})
        self.error_on_dryrun()
        key_pair = self.ec2_backend.create_key_pair(key_name, key_type, tags)
        return ActionResult(key_pair)

    def delete_key_pair(self) -> ActionResult:
        key_name = self._get_param("KeyName")
        self.error_on_dryrun()
        key_pair = self.ec2_backend.delete_key_pair(key_name)
        result = {"Return": True, "KeyPairId": key_pair.id if key_pair else None}
        return ActionResult(result)

    def describe_key_pairs(self) -> ActionResult:
        names = self._get_param("KeyNames", [])
        filters = self._filters_from_querystring()
        key_pairs = self.ec2_backend.describe_key_pairs(names, filters)
        include_public_key = self._get_bool_param("IncludePublicKey", False)
        public_key_attribute_path = (
            "DescribeKeyPairsResult.KeyPairs.KeyPairInfo.PublicKey"
        )
        if include_public_key:
            self._include_in_response(public_key_attribute_path)
        else:
            self._exclude_from_response(public_key_attribute_path)
        result = {"KeyPairs": key_pairs}
        return ActionResult(result)

    def import_key_pair(self) -> ActionResult:
        key_name = self._get_param("KeyName")
        material = self._get_param("PublicKeyMaterial")
        tags = self._parse_tag_specification("key-pair").get("key-pair", {})
        self.error_on_dryrun()
        key_pair = self.ec2_backend.import_key_pair(key_name, material, tags)
        return ActionResult(key_pair)



"""Handles incoming codedeploy requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import CodeDeployBackend, codedeploy_backends


class CodeDeployResponse(BaseResponse):
    """Handler for CodeDeploy requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="codedeploy")
        self.default_response_headers = {"Content-Type": "application/json"}

    @property
    def codedeploy_backend(self) -> CodeDeployBackend:
        """Return backend instance specific for this region."""
        return codedeploy_backends[self.current_account][self.region]

    def batch_get_applications(self) -> str:
        application_names = self._get_param("applicationNames")
        applications = self.codedeploy_backend.batch_get_applications(
            application_names=application_names,
        )

        applications_info = {
            "applicationsInfo": [app.to_dict() for app in applications]
        }
        return json.dumps(applications_info)

    def get_application(self) -> str:
        application_name = self._get_param("applicationName")
        application = self.codedeploy_backend.get_application(
            application_name=application_name,
        )

        return json.dumps({"application": application.to_dict()})

    def get_deployment(self) -> str:
        deployment_id = self._get_param("deploymentId")
        deployment = self.codedeploy_backend.get_deployment(
            deployment_id=deployment_id,
        )
        return json.dumps({"deploymentInfo": deployment.to_dict()})

    def get_deployment_group(self) -> str:
        application_name = self._get_param("applicationName")
        deployment_group_name = self._get_param("deploymentGroupName")
        deployment_group = self.codedeploy_backend.get_deployment_group(
            application_name=application_name,
            deployment_group_name=deployment_group_name,
        )
        return json.dumps({"deploymentGroupInfo": deployment_group.to_dict()})

    def batch_get_deployments(self) -> str:
        deployment_ids = self._get_param("deploymentIds")
        deployments = self.codedeploy_backend.batch_get_deployments(
            deployment_ids=deployment_ids,
        )

        deployments_info = {
            "deploymentsInfo": [deployment.to_dict() for deployment in deployments]
        }
        return json.dumps(deployments_info)

    def create_application(self) -> str:
        application_name = self._get_param("applicationName")
        compute_platform = self._get_param("computePlatform")
        tags = self._get_param("tags")
        application_id = self.codedeploy_backend.create_application(
            application_name=application_name,
            compute_platform=compute_platform,
            tags=tags,
        )
        return json.dumps({"applicationId": application_id})

    def create_deployment(self) -> str:
        application_name = self._get_param("applicationName")
        deployment_group_name = self._get_param("deploymentGroupName")
        revision = self._get_param("revision")
        deployment_config_name = self._get_param("deploymentConfigName")
        description = self._get_param("description")
        ignore_application_stop_failures = self._get_bool_param(
            "ignoreApplicationStopFailures"
        )
        target_instances = self._get_param("targetInstances")
        auto_rollback_configuration = self._get_param("autoRollbackConfiguration")
        update_outdated_instances_only = self._get_param("updateOutdatedInstancesOnly")
        file_exists_behavior = self._get_param("fileExistsBehavior")
        override_alarm_configuration = self._get_param("overrideAlarmConfiguration")
        deployment_id = self.codedeploy_backend.create_deployment(
            application_name=application_name,
            deployment_group_name=deployment_group_name,
            revision=revision,
            deployment_config_name=deployment_config_name,
            description=description,
            ignore_application_stop_failures=ignore_application_stop_failures,
            target_instances=target_instances,
            auto_rollback_configuration=auto_rollback_configuration,
            update_outdated_instances_only=update_outdated_instances_only,
            file_exists_behavior=file_exists_behavior,
            override_alarm_configuration=override_alarm_configuration,
        )
        return json.dumps({"deploymentId": deployment_id})

    def create_deployment_group(self) -> str:
        application_name = self._get_param("applicationName")
        deployment_group_name = self._get_param("deploymentGroupName")
        deployment_config_name = self._get_param("deploymentConfigName")
        ec2_tag_filters = self._get_param("ec2TagFilters")
        on_premises_instance_tag_filters = self._get_param(
            "onPremisesInstanceTagFilters"
        )
        auto_scaling_groups = self._get_param("autoScalingGroups")
        service_role_arn = self._get_param("serviceRoleArn")
        trigger_configurations = self._get_param("triggerConfigurations")
        alarm_configuration = self._get_param("alarmConfiguration")
        auto_rollback_configuration = self._get_param("autoRollbackConfiguration")
        outdated_instances_strategy = self._get_param("outdatedInstancesStrategy")
        deployment_style = self._get_param("deploymentStyle")
        blue_green_deployment_configuration = self._get_param(
            "blueGreenDeploymentConfiguration"
        )
        load_balancer_info = self._get_param("loadBalancerInfo")
        ec2_tag_set = self._get_param("ec2TagSet")
        ecs_services = self._get_param("ecsServices")
        on_premises_tag_set = self._get_param("onPremisesTagSet")
        tags = self._get_param("tags")
        termination_hook_enabled = self._get_param("terminationHookEnabled")
        deployment_group_id = self.codedeploy_backend.create_deployment_group(
            application_name=application_name,
            deployment_group_name=deployment_group_name,
            deployment_config_name=deployment_config_name,
            ec2_tag_filters=ec2_tag_filters,
            on_premises_instance_tag_filters=on_premises_instance_tag_filters,
            auto_scaling_groups=auto_scaling_groups,
            service_role_arn=service_role_arn,
            trigger_configurations=trigger_configurations,
            alarm_configuration=alarm_configuration,
            auto_rollback_configuration=auto_rollback_configuration,
            outdated_instances_strategy=outdated_instances_strategy,
            deployment_style=deployment_style,
            blue_green_deployment_configuration=blue_green_deployment_configuration,
            load_balancer_info=load_balancer_info,
            ec2_tag_set=ec2_tag_set,
            ecs_services=ecs_services,
            on_premises_tag_set=on_premises_tag_set,
            tags=tags,
            termination_hook_enabled=termination_hook_enabled,
        )
        return json.dumps({"deploymentGroupId": deployment_group_id})

    def list_applications(self) -> str:
        applications = self.codedeploy_backend.list_applications()
        return json.dumps({"applications": applications})

    def list_deployments(self) -> str:
        application_name = self._get_param("applicationName")
        deployment_group_name = self._get_param("deploymentGroupName")
        external_id = self._get_param("externalId")
        include_only_statuses = self._get_param("includeOnlyStatuses")
        create_time_range = self._get_param("createTimeRange")
        deployments = self.codedeploy_backend.list_deployments(
            application_name=application_name,
            deployment_group_name=deployment_group_name,
            external_id=external_id,
            include_only_statuses=include_only_statuses,
            create_time_range=create_time_range,
        )
        return json.dumps({"deployments": deployments})

    def list_deployment_groups(self) -> str:
        application_name = self._get_param("applicationName")
        next_token = self._get_param("nextToken", "")
        deployment_groups = self.codedeploy_backend.list_deployment_groups(
            application_name=application_name,
            next_token=next_token,
        )
        return json.dumps(
            {
                "applicationName": application_name,
                "deploymentGroups": deployment_groups,
                "nextToken": next_token,
            }
        )

    def list_tags_for_resource(self) -> str:
        """Handler for list_tags_for_resource API call."""
        resource_arn = self._get_param("ResourceArn")
        tags_response = self.codedeploy_backend.list_tags_for_resource(resource_arn)
        return json.dumps(tags_response)

    def tag_resource(self) -> str:
        """Handler for tag_resource API call."""
        resource_arn = self._get_param("ResourceArn")
        tags = self._get_param("Tags")
        response = self.codedeploy_backend.tag_resource(resource_arn, tags)
        return json.dumps(response)

    def untag_resource(self) -> str:
        """Handler for untag_resource API call."""
        resource_arn = self._get_param("ResourceArn")
        tag_keys = self._get_param("TagKeys")
        response = self.codedeploy_backend.untag_resource(resource_arn, tag_keys)
        return json.dumps(response)



from __future__ import annotations

import base64
import binascii


class Base64EncodedString:
    """
    A type to encapsulate a Base64 encoded string.
    """

    def __init__(self, encoded_string: str):
        if not isinstance(encoded_string, str):
            raise TypeError("Base64EncodedString must be initialized with a string.")
        try:
            base64.b64decode(encoded_string, validate=True)
        except binascii.Error:
            raise ValueError("Invalid Base64 encoded string provided.")
        self._encoded_string = encoded_string

    def as_bytes(self, encoding: str = "utf-8") -> bytes:
        return self._encoded_string.encode(encoding)

    def decode(self, encoding: str = "utf-8") -> str:
        decoded_bytes = base64.b64decode(self._encoded_string)
        return decoded_bytes.decode(encoding)

    @classmethod
    def from_encoded_bytes(cls, raw_bytes: bytes) -> Base64EncodedString:
        return cls(raw_bytes.decode("ascii"))

    @classmethod
    def from_raw_string(
        cls, raw_string: str, encoding: str = "utf-8"
    ) -> Base64EncodedString:
        encoded_bytes = base64.b64encode(raw_string.encode(encoding))
        return cls.from_encoded_bytes(encoded_bytes)

    def __str__(self) -> str:
        return self._encoded_string

    def __repr__(self) -> str:
        return f"Base64EncodedString('{self._encoded_string}')"



import ast
from typing import Optional

from antlr4 import ParserRuleContext
from antlr4.tree.Tree import ParseTree, TerminalNodeImpl


def is_production(
    pt: ParseTree, rule_index: Optional[int] = None
) -> Optional[ParserRuleContext]:
    if isinstance(pt, ParserRuleContext):
        prc = pt.getRuleContext()  # noqa
        if rule_index is not None:
            return prc if prc.getRuleIndex() == rule_index else None
        return prc
    return None


def is_terminal(
    pt: ParseTree, token_type: Optional[int] = None
) -> Optional[TerminalNodeImpl]:
    if isinstance(pt, TerminalNodeImpl):
        if token_type is not None:
            return pt if pt.getSymbol().type == token_type else None
        return pt
    return None


def from_string_literal(parser_rule_context: ParserRuleContext) -> Optional[str]:
    string_literal = parser_rule_context.getText()
    if string_literal.startswith('"') and string_literal.endswith('"'):
        string_literal = string_literal[1:-1]
    # Interpret escape sequences into their character representations
    try:
        string_literal = ast.literal_eval(f'"{string_literal}"')
    except Exception:
        # Fallback if literal_eval fails
        pass
    return string_literal



from typing import Any, Optional

from moto.core.responses import ActionResult, BaseResponse, EmptyResult, PaginatedResult
from moto.core.utils import utcnow

from .exceptions import ValidationError
from .models import SESBackend, ses_backends


class EmailResponse(BaseResponse):
    def __init__(self) -> None:
        super().__init__(service_name="ses")
        self.automated_parameter_parsing = True

    @property
    def backend(self) -> SESBackend:
        return ses_backends[self.current_account][self.region]

    def verify_email_identity(self) -> ActionResult:
        address = self._get_param("EmailAddress")
        self.backend.verify_email_identity(address)
        return EmptyResult()

    def verify_email_address(self) -> ActionResult:
        address = self._get_param("EmailAddress")
        self.backend.verify_email_address(address)
        return EmptyResult()

    def list_identities(self) -> ActionResult:
        identity_type = self._get_param("IdentityType")
        if identity_type not in [None, "EmailAddress", "Domain"]:
            raise ValidationError(
                f"Value '{identity_type}' at 'identityType' failed to satisfy constraint: Member must satisfy enum value set: [Domain, EmailAddress]"
            )
        identities = self.backend.list_identities(identity_type)
        result = {"Identities": identities}
        return ActionResult(result)

    def list_verified_email_addresses(self) -> ActionResult:
        email_addresses = self.backend.list_verified_email_addresses()
        result = {"VerifiedEmailAddresses": email_addresses}
        return ActionResult(result)

    def verify_domain_dkim(self) -> ActionResult:
        domain = self._get_param("Domain")
        self.backend.verify_domain(domain)
        result = {
            "DkimTokens": [
                "vvjuipp74whm76gqoni7qmwwn4w4qusjiainivf6sf",
                "3frqe7jn4obpuxjpwpolz6ipb3k5nvt2nhjpik2oy",
                "wrqplteh7oodxnad7hsl4mixg2uavzneazxv5sxi2",
            ]
        }
        return ActionResult(result)

    def verify_domain_identity(self) -> ActionResult:
        domain = self._get_param("Domain")
        self.backend.verify_domain(domain)
        result = {"VerificationToken": "QTKknzFg2J4ygwa+XvHAxUl1hyHoY0gVfZdfjIedHZ0="}
        return ActionResult(result)

    def delete_identity(self) -> ActionResult:
        domain = self._get_param("Identity")
        self.backend.delete_identity(domain)
        return EmptyResult()

    def send_email(self) -> ActionResult:
        bodydatakey = "Message.Body.Text.Data"
        if "Message.Body.Html.Data" in self.querystring:
            bodydatakey = "Message.Body.Html.Data"
        body = self._get_param(bodydatakey)
        source = self._get_param("Source")
        subject = self._get_param("Message.Subject.Data")
        destinations = self._get_param("Destination", {})
        message = self.backend.send_email(source, subject, body, destinations)
        result = {"MessageId": message.id}
        return ActionResult(result)

    def send_templated_email(self) -> ActionResult:
        source = self._get_param("Source")
        template = self._get_param("Template")
        template_data = self._get_param("TemplateData")
        destinations = self._get_param("Destination", {})
        message = self.backend.send_templated_email(
            source, template, template_data, destinations
        )
        result = {"MessageId": message.id}
        return ActionResult(result)

    def send_bulk_templated_email(self) -> ActionResult:
        source = self._get_param("Source")
        template = self._get_param("Template")
        template_data = self._get_param("DefaultTemplateData")
        destinations = self._get_param("Destinations", [])
        message = self.backend.send_bulk_templated_email(
            source,
            template,
            template_data,
            destinations,
        )
        result = {"Status": [{"MessageId": msg_id} for msg_id in message.ids]}
        return ActionResult(result)

    def send_raw_email(self) -> ActionResult:
        source = self._get_param("Source")
        raw_data = self._get_param("RawMessage.Data")
        raw_data = raw_data.decode("utf-8")
        destinations = self._get_param("Destinations", [])
        message = self.backend.send_raw_email(source, destinations, raw_data)  # type: ignore[arg-type]
        result = {"MessageId": message.id}
        return ActionResult(result)

    def get_send_quota(self) -> ActionResult:
        quota = self.backend.get_send_quota()
        return ActionResult(quota)

    def get_identity_notification_attributes(self) -> ActionResult:
        identities = self._get_param("Identities")
        identities = self.backend.get_identity_notification_attributes(identities)
        result = {"NotificationAttributes": identities}
        return ActionResult(result)

    def set_identity_feedback_forwarding_enabled(self) -> ActionResult:
        identity = self._get_param("Identity")
        enabled = self._get_bool_param("ForwardingEnabled", False)
        self.backend.set_identity_feedback_forwarding_enabled(identity, enabled)
        return EmptyResult()

    def set_identity_notification_topic(self) -> ActionResult:
        identity = self._get_param("Identity")
        not_type = self._get_param("NotificationType")
        sns_topic = self._get_param("SnsTopic")
        self.backend.set_identity_notification_topic(identity, not_type, sns_topic)
        return EmptyResult()

    def get_send_statistics(self) -> ActionResult:
        statistics = self.backend.get_send_statistics()
        result = {"SendDataPoints": [statistics]}
        return ActionResult(result)

    def create_configuration_set(self) -> ActionResult:
        configuration_set_name = self._get_param("ConfigurationSet.Name")
        self.backend.create_configuration_set(
            configuration_set_name=configuration_set_name
        )
        return EmptyResult()

    def describe_configuration_set(self) -> ActionResult:
        configuration_set_name = self._get_param("ConfigurationSetName")
        config_set = self.backend.describe_configuration_set(configuration_set_name)

        attribute_names = self._get_param("ConfigurationSetAttributeNames", [])

        event_destination: Optional[dict[str, Any]] = None
        if "eventDestinations" in attribute_names:
            event_destination = self.backend.config_set_event_destination.get(
                configuration_set_name
            )

        result = {
            "ConfigurationSet": {"Name": config_set.configuration_set_name},
            "EventDestinations": [event_destination],
        }
        return ActionResult(result)

    def create_configuration_set_event_destination(self) -> ActionResult:
        configuration_set_name = self._get_param("ConfigurationSetName")
        event_destination = self._get_param("EventDestination", {})
        self.backend.create_configuration_set_event_destination(
            configuration_set_name=configuration_set_name,
            event_destination=event_destination,
        )
        return EmptyResult()

    def create_template(self) -> ActionResult:
        template_data = self._get_param("Template", {})
        template_info = {}
        template_info["text_part"] = template_data.get("TextPart", "")
        template_info["html_part"] = template_data.get("HtmlPart", "")
        template_info["template_name"] = template_data.get("TemplateName", "")
        template_info["subject_part"] = template_data.get("SubjectPart", "")
        template_info["Timestamp"] = utcnow()
        self.backend.add_template(template_info=template_info)
        return EmptyResult()

    def update_template(self) -> ActionResult:
        template_data = self._get_param("Template", {})
        template_info = {}
        template_info["text_part"] = template_data.get("TextPart", "")
        template_info["html_part"] = template_data.get("HtmlPart", "")
        template_info["template_name"] = template_data.get("TemplateName", "")
        template_info["subject_part"] = template_data.get("SubjectPart", "")
        template_info["Timestamp"] = utcnow()
        self.backend.update_template(template_info=template_info)
        return EmptyResult()

    def get_template(self) -> ActionResult:
        template_name = self._get_param("TemplateName")
        template_data = self.backend.get_template(template_name)
        result = {"Template": template_data}
        return ActionResult(result)

    def list_templates(self) -> ActionResult:
        email_templates = self.backend.list_templates()
        metadata = [
            {"Name": t["template_name"], "CreatedTimestamp": t["Timestamp"]}
            for t in email_templates
        ]
        result = {"TemplatesMetadata": metadata}
        return ActionResult(result)

    def test_render_template(self) -> ActionResult:
        render_info = {
            "name": self._get_param("TemplateName"),
            "data": self._get_param("TemplateData"),
        }
        rendered_template = self.backend.render_template(render_info)
        result = {"RenderedTemplate": rendered_template}
        return ActionResult(result)

    def delete_template(self) -> ActionResult:
        name = self._get_param("TemplateName")
        self.backend.delete_template(name)
        return EmptyResult()

    def create_receipt_rule_set(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName")
        self.backend.create_receipt_rule_set(rule_set_name)
        return EmptyResult()

    def create_receipt_rule(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName", "")
        rule = self._get_param("Rule", {})
        after = self._get_param("After", "")
        self.backend.create_receipt_rule(rule_set_name, rule, after)
        return EmptyResult()

    def clone_receipt_rule_set(self) -> ActionResult:
        original_rule_set_name = self._get_param("OriginalRuleSetName")
        rule_set_name = self._get_param("RuleSetName")
        self.backend.clone_receipt_rule_set(original_rule_set_name, rule_set_name)
        return EmptyResult()

    def set_active_receipt_rule_set(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName")
        result = self.backend.set_active_receipt_rule_set(rule_set_name)
        return ActionResult(result)

    def describe_receipt_rule_set(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName")
        rule_set = self.backend.describe_receipt_rule_set(rule_set_name)
        return ActionResult(rule_set)

    def describe_active_receipt_rule_set(self) -> ActionResult:
        rule_set = self.backend.describe_active_receipt_rule_set()
        return ActionResult(rule_set)

    def delete_receipt_rule_set(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName")
        self.backend.delete_receipt_rule_set(rule_set_name)
        return EmptyResult()

    def list_receipt_rule_sets(self) -> ActionResult:
        rule_sets = self.backend.list_receipt_rule_sets()
        result = {"RuleSets": [rs.metadata for rs in rule_sets]}
        return ActionResult(result)

    def describe_receipt_rule(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName")
        rule_name = self._get_param("RuleName")
        receipt_rule = self.backend.describe_receipt_rule(rule_set_name, rule_name)
        result = {"Rule": receipt_rule}
        return ActionResult(result)

    def update_receipt_rule(self) -> ActionResult:
        rule_set_name = self._get_param("RuleSetName", "")
        rule = self._get_param("Rule", {})
        self.backend.update_receipt_rule(rule_set_name, rule)
        return EmptyResult()

    def set_identity_mail_from_domain(self) -> ActionResult:
        identity = self._get_param("Identity")
        mail_from_domain = self._get_param("MailFromDomain")
        behavior_on_mx_failure = self._get_param("BehaviorOnMXFailure")

        self.backend.set_identity_mail_from_domain(
            identity, mail_from_domain, behavior_on_mx_failure
        )
        return EmptyResult()

    def get_identity_mail_from_domain_attributes(self) -> ActionResult:
        identities = self._get_param("Identities", [])
        attributes_by_identity = self.backend.get_identity_mail_from_domain_attributes(
            identities
        )
        result = {"MailFromDomainAttributes": attributes_by_identity}
        return ActionResult(result)

    def get_identity_verification_attributes(self) -> ActionResult:
        identities = self._get_param("Identities", [])
        verification_attributes = self.backend.get_identity_verification_attributes(
            identities=identities,
        )
        result = {"VerificationAttributes": verification_attributes}
        return ActionResult(result)

    def delete_configuration_set(self) -> ActionResult:
        configuration_set_name = self._get_param("ConfigurationSetName")
        self.backend.delete_configuration_set(
            configuration_set_name=str(configuration_set_name)
        )
        return EmptyResult()

    def list_configuration_sets(self) -> ActionResult:
        configuration_sets = self.backend.list_configuration_sets()
        config_set_names = [c.configuration_set_name for c in configuration_sets]
        result = {
            "ConfigurationSets": [{"Name": name} for name in config_set_names],
        }
        return PaginatedResult(result)

    def update_configuration_set_reputation_metrics_enabled(self) -> ActionResult:
        configuration_set_name = self._get_param("ConfigurationSetName")
        enabled = self._get_bool_param("Enabled", False)
        self.backend.update_configuration_set_reputation_metrics_enabled(
            configuration_set_name=configuration_set_name,
            enabled=enabled,
        )
        return EmptyResult()

    def get_identity_dkim_attributes(self) -> ActionResult:
        identities = self._get_param("Identities", [])
        dkim_attributes = self.backend.get_identity_dkim_attributes(identities)
        result = {"DkimAttributes": dkim_attributes}
        return ActionResult(result)



from collections import OrderedDict
from typing import Final

from moto.stepfunctions.parser.api import (
    StateName,
    VariableName,
    VariableNameList,
    VariableReferences,
)
from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.jsonata.jsonata import (
    VariableReference,
    extract_jsonata_variable_references,
)
from moto.stepfunctions.parser.asl.static_analyser.intrinsic.variable_names_intrinsic_static_analyser import (
    VariableNamesIntrinsicStaticAnalyser,
)
from moto.stepfunctions.parser.asl.static_analyser.static_analyser import StaticAnalyser


class VariableReferencesStaticAnalyser(StaticAnalyser):
    @staticmethod
    def process_and_get(definition: str) -> VariableReferences:
        analyser = VariableReferencesStaticAnalyser()
        analyser.analyse(definition=definition)
        return analyser.get_variable_references()

    _fringe_state_names: Final[list[StateName]]
    _variable_references: Final[VariableReferences]

    def __init__(self):
        super().__init__()
        self._fringe_state_names = []
        self._variable_references = OrderedDict()

    def get_variable_references(self) -> VariableReferences:
        return self._variable_references

    def _enter_state(self, state_name: StateName) -> None:
        self._fringe_state_names.append(state_name)

    def _exit_state(self) -> None:
        self._fringe_state_names.pop()

    def visitState_decl(self, ctx: ASLParser.State_declContext) -> None:
        state_name: str = ctx.string_literal().getText()[1:-1]
        self._enter_state(state_name=state_name)
        super().visitState_decl(ctx=ctx)
        self._exit_state()

    def _put_variable_reference(self, variable_reference: VariableReference) -> None:
        variable_name: VariableName = variable_reference[1:]
        self._put_variable_name(variable_name)

    def _put_variable_name(self, variable_name: VariableName) -> None:
        state_name = self._fringe_state_names[-1]
        variable_name_list: VariableNameList = self._variable_references.get(
            state_name, []
        )
        if variable_name in variable_name_list:
            return
        variable_name_list.append(variable_name)
        if state_name not in self._variable_references:
            self._variable_references[state_name] = variable_name_list

    def visitString_variable_sample(self, ctx: ASLParser.String_variable_sampleContext):
        reference_body = ctx.getText()[1:-1]
        variable_references: set[VariableReference] = (
            extract_jsonata_variable_references(reference_body)
        )
        for variable_reference in variable_references:
            self._put_variable_reference(variable_reference)

    def visitString_intrinsic_function(
        self, ctx: ASLParser.String_intrinsic_functionContext
    ):
        definition_body = ctx.getText()[1:-1]
        variable_name_list: VariableNameList = (
            VariableNamesIntrinsicStaticAnalyser.process_and_get(definition_body)
        )
        for variable_name in variable_name_list:
            self._put_variable_name(variable_name)

    def visitString_literal(self, ctx: ASLParser.String_literalContext):
        # Prune everything parsed as a string literal.
        return



import copy
import random
import string
from re import compile as re_compile
from typing import Any, Optional, Union

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import utcnow
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition

from ..moto_api._internal import mock_random
from .exceptions import (
    CacheClusterAlreadyExists,
    CacheClusterNotFound,
    CacheSubnetGroupAlreadyExists,
    CacheSubnetGroupNotFound,
    InvalidARNFault,
    InvalidParameterCombinationException,
    InvalidParameterValueException,
    InvalidSubnet,
    ReplicationGroupAlreadyExists,
    ReplicationGroupNotFound,
    SnapshotAlreadyExists,
    SnapshotNotFound,
    UserAlreadyExists,
    UserNotFound,
)
from .utils import AuthenticationTypes


class User(BaseModel):
    def __init__(
        self,
        account_id: str,
        region: str,
        user_id: str,
        user_name: str,
        access_string: str,
        engine: str,
        no_password_required: bool,
        passwords: Optional[list[str]] = None,
        authentication_type: Optional[str] = None,
        tags: Optional[list[dict[str, str]]] = None,
    ):
        self.id = user_id
        self.name = user_name
        self.engine = engine

        self.passwords = passwords or []
        self.access_string = access_string
        self.no_password_required = no_password_required
        self.status = "active"
        self.minimum_engine_version = "6.0"
        self.user_group_ids: list[str] = []
        self.region = region
        self.arn = f"arn:{get_partition(self.region)}:elasticache:{self.region}:{account_id}:user:{self.id}"
        self.authentication_type = authentication_type
        self.tags = tags or []

    @property
    def authentication(self) -> dict[str, Union[str, int, None]]:
        return {
            "Type": "no-password"
            if self.no_password_required
            else self.authentication_type,
            "PasswordCount": len(self.passwords) if self.passwords else None,
        }


class CacheCluster(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        cache_cluster_id: str,
        cache_node_type: str,
        replication_group_id: Optional[str],
        az_mode: Optional[str],
        preferred_availability_zone: Optional[str],
        num_cache_nodes: Optional[int],
        engine: Optional[str],
        engine_version: Optional[str],
        cache_parameter_group_name: Optional[str],
        cache_subnet_group_name: Optional[str],
        transit_encryption_enabled: Optional[bool],
        network_type: Optional[str],
        ip_discovery: Optional[str],
        snapshot_name: Optional[str],
        preferred_maintenance_window: Optional[str],
        port: Optional[int],
        notification_topic_arn: Optional[str],
        auto_minor_version_upgrade: Optional[bool],
        snapshot_retention_limit: Optional[int],
        snapshot_window: Optional[str],
        auth_token: Optional[str],
        outpost_mode: Optional[str],
        preferred_outpost_arn: Optional[str],
        preferred_availability_zones: Optional[list[str]],
        cache_security_group_names: Optional[list[str]],
        security_group_ids: Optional[list[str]],
        tags: Optional[list[dict[str, str]]],
        snapshot_arns: Optional[list[str]],
        preferred_outpost_arns: Optional[list[str]],
        log_delivery_configurations: list[dict[str, Any]],
        cache_node_ids_to_remove: Optional[list[str]],
        cache_node_ids_to_reboot: Optional[list[str]],
    ):
        self.cache_cluster_id = cache_cluster_id
        self.az_mode = az_mode
        self.preferred_availability_zone = preferred_availability_zone
        self.preferred_availability_zones = preferred_availability_zones or []
        self.engine = engine or "redis"
        self.engine_version = engine_version
        if engine == "redis":
            self.num_cache_nodes = 1
            self.replication_group_id = replication_group_id
            self.snapshot_arns = snapshot_arns or []
            self.snapshot_name = snapshot_name
            self.snapshot_window = snapshot_window
        if engine == "memcached":
            if num_cache_nodes is None:
                self.num_cache_nodes = 1
            elif 1 <= num_cache_nodes <= 40:
                self.num_cache_nodes = num_cache_nodes
        self.cache_node_type = cache_node_type
        self.cache_parameter_group_name = cache_parameter_group_name
        self.cache_subnet_group_name = cache_subnet_group_name
        self.cache_security_group_names = cache_security_group_names or []
        self.security_group_ids = security_group_ids or []
        self.tags = tags or []
        self.preferred_maintenance_window = preferred_maintenance_window
        self.port = port or 6379
        self.notification_topic_arn = notification_topic_arn
        self.auto_minor_version_upgrade = auto_minor_version_upgrade
        self.snapshot_retention_limit = snapshot_retention_limit or 0
        self.auth_token = auth_token
        self.auth_token_enabled = bool(auth_token)
        self.outpost_mode = outpost_mode
        self.preferred_outpost_arn = preferred_outpost_arn
        self.preferred_outpost_arns = preferred_outpost_arns or []
        self.log_delivery_configurations = log_delivery_configurations or []
        self.transit_encryption_enabled = transit_encryption_enabled
        self.network_type = network_type
        self.ip_discovery = ip_discovery
        self.cache_node_ids_to_remove = cache_node_ids_to_remove
        self.cache_node_ids_to_reboot = cache_node_ids_to_reboot

        self.cache_cluster_create_time = utcnow()
        self.auth_token_last_modified_date = utcnow()
        self.cache_cluster_status = "available"
        self.arn = f"arn:{get_partition(region_name)}:elasticache:{region_name}:{account_id}:cluster:{cache_cluster_id}"
        self.cache_node_id = str(mock_random.uuid4())
        self.status = "available"


class CacheSubnetGroup(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        cache_subnet_group_name: str,
        cache_subnet_group_description: str,
        subnet_ids: list[str],
        tags: Optional[list[dict[str, str]]],
    ):
        self.cache_subnet_group_name = cache_subnet_group_name
        self.cache_subnet_group_description = cache_subnet_group_description
        self.subnet_ids = subnet_ids
        self.tags = tags or []

        # Only import ec2_backends if necessary
        from moto.ec2.models import ec2_backends

        ec2_backend = ec2_backends[account_id][region_name]
        self.supported_network_types = []
        self.subnets = []
        vpc_exists = False
        try:
            # Get VPC details from provided subnet IDs
            subnets = ec2_backend.describe_subnets(subnet_ids=subnet_ids)
            vpc_exists = True
        except Exception as e:
            # Should raise InvalidSubnet if subnet_ids are invalid
            if "InvalidSubnet" in str(e):
                for subnet_id in subnet_ids:
                    subnet_response: dict[str, Any] = {}
                    subnet_response["subnet_identifier"] = subnet_id
                    subnet_response["Subnet_availability_zone"] = {"Name": "us-east-1a"}
                    subnet_response["supported_network_types"] = ["ipv4"]
                    self.subnets.append(subnet_response)
                vpcs = ["vpc-0123456789abcdef0"]
                self.supported_network_types = ["ipv4"]

        if vpc_exists:
            vpcs = []
            for subnet in subnets:
                subnet_response = {}
                vpcs.append(subnet.vpc_id)
                subnet_response["subnet_identifier"] = subnet.id
                subnet_response["subnet_availability_zone"] = subnet.availability_zone
                subnet_response["supported_network_types"] = []
                if subnet.vpc_id != vpcs[0]:
                    raise InvalidSubnet(subnet_id=subnet.id)

                # ipv6 native subnets only appends ipv6
                # You can't mix ipv6 native subnets with other types of subnets
                if subnet.ipv6_native:
                    self.supported_network_types.append("ipv6")
                    subnet_response["supported_network_types"].append("ipv6")

                # ipv4 only and dual_stack subnets both append ipv4
                elif subnet.cidr_block:
                    self.supported_network_types.append("ipv4")
                    subnet_response["supported_network_types"].append("ipv4")

                if subnet.ipv6_cidr_block_associations and subnet.cidr_block:
                    self.supported_network_types.append("dual_stack")
                    subnet_response["supported_network_types"].append("dual_stack")

                self.subnets.append(subnet_response)

            if self.supported_network_types:
                self.supported_network_types = list(set(self.supported_network_types))

        self.arn = f"arn:aws:elasticache:{region_name}:{account_id}:subnetgroup:{cache_subnet_group_name}"
        self.vpc_id = vpcs[0] if vpcs else None


class ReplicationGroup(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        replication_group_id: str,
        preferred_cache_cluster_azs: Optional[list[str]],
        num_cache_clusters: Optional[int],
        num_node_groups: Optional[int],
        replicas_per_node_group: Optional[int],
        node_group_configuration: list[dict[str, Any]],
        preferred_maintenance_window: Optional[str],
        replication_group_description: str,
        primary_cluster_id: Optional[str],
        automatic_failover_enabled: bool,
        global_replication_group_id: Optional[str],
        multi_az_enabled: Optional[bool],
        port: Optional[int],
        snapshot_retention_limit: int,
        snapshot_window: Optional[str],
        cluster_mode: Optional[str],
        cache_node_type: Optional[str],
        auth_token: Optional[str],
        transit_encryption_enabled: Optional[bool],
        at_rest_encryption_enabled: Optional[bool],
        kms_key_id: Optional[str],
        user_group_ids: list[str],
        log_delivery_configurations: list[dict[str, Any]],
        data_tiering_enabled: Optional[bool],
        auto_minor_version_upgrade: Optional[bool],
        engine: Optional[str],
        network_type: Optional[str],
        ip_discovery: Optional[str],
        transit_encryption_mode: Optional[str],
        cache_security_group_names: Optional[list[str]],
        cache_subnet_group_name: Optional[str],
        security_group_ids: Optional[list[str]],
        tags: Optional[list[dict[str, str]]],
        notification_topic_arn: Optional[str],
        serverless_cache_snapshot_name: Optional[str],
        cache_parameter_group_name: Optional[str],
        engine_version: Optional[str],
        snapshot_arns: Optional[list[str]],
        snapshot_name: Optional[str],
    ):
        tags = tags or []
        self.cluster_mode = cluster_mode or "disabled"
        self.replication_group_id = replication_group_id
        self.description = replication_group_description
        self.port = port or 6379
        self.status = "available"

        preferred_cache_cluster_azs = preferred_cache_cluster_azs or []
        replicas_per_node_group = replicas_per_node_group or 0

        # This field is only present if we are creating a
        # secondary replication group and associating it with an existing
        # global replication group and primary cluster.
        if global_replication_group_id:
            self.global_replication_group_id = global_replication_group_id
            self.global_replication_group_member_role = "SECONDARY"
        else:
            self.global_replication_group_id = ""

        # Leaving off PendingModifiedValues for now
        # until global_replication_group is implemented.
        self.primary_cluster_id = primary_cluster_id

        self.node_groups = []
        random_str = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=6)
        )
        region_short = "use1"
        replication_group_domain = (
            f"{replication_group_id}.{random_str}.{region_short}.cache.amazonaws.com"
        )
        self.member_clusters: list[str] = []
        self.member_clusters_outpost_arns: list[str] = []
        if not num_node_groups:
            num_node_groups = 1

        for i in range(num_node_groups):
            node_group: dict[str, Any] = {}
            node_group["node_group_id"] = f"{i + 1:0>4}"
            node_group["status"] = "available"
            node_group["node_group_members"] = []

            if self.cluster_mode == "enabled":
                if i == 0:
                    node_group["slots"] = "0-5461"
                else:
                    # slots cannot overlap among node groups
                    node_group["slots"] = (
                        f"{(5460 * (i)) + (i + 1)}-{(5460 * (i + 1)) + (i + 1)}"
                    )

                replicas_per_node_group = replicas_per_node_group or 0
                self._set_node_members_clusters_enabled(
                    member_clusters=self.member_clusters,
                    replication_group_id=replication_group_id,
                    replicas_per_node_group=replicas_per_node_group,
                    node_group_configuration=node_group_configuration,
                    node_group=node_group,
                    member_clusters_outpost_arns=self.member_clusters_outpost_arns,
                )

                self.configuration_endpoint: dict[str, Any] = {}
                self.configuration_endpoint["address"] = (
                    f"clustercfg.{replication_group_domain}"
                )
                self.configuration_endpoint["port"] = self.port

            # self.cluster_mode is disabled or not set
            else:
                node_group["primary_endpoint_address"] = (
                    f"master.{replication_group_domain}"
                )
                node_group["reader_endpoint_address"] = (
                    f"replica.{replication_group_domain}"
                )
                node_group["port"] = self.port
                node_group["primary_endpoint"] = {
                    "address": node_group["primary_endpoint_address"],
                    "port": self.port,
                }

                num_cache_clusters = num_cache_clusters or 1

                self._set_node_members_clusters_disabled(
                    member_clusters=self.member_clusters,
                    replication_group_id=replication_group_id,
                    replication_group_domain=replication_group_domain,
                    node_group=node_group,
                    preferred_cache_cluster_azs=preferred_cache_cluster_azs,
                    num_cache_clusters=num_cache_clusters,
                    replicas_per_node_group=replicas_per_node_group,
                )
            self.node_groups.append(node_group)

        self.automatic_failover = (
            "enabled" if automatic_failover_enabled else "disabled"
        )
        self.multi_az = "enabled" if multi_az_enabled else "disabled"
        self.snapshot_retention_limit = snapshot_retention_limit or 0
        self.snapshot_window = snapshot_window or "05:00-6:00"
        self.snapshotting_cluster_id = None
        if self.snapshot_retention_limit > 0:
            if len(self.member_clusters) > 1:
                self.snapshotting_cluster_id = self.member_clusters[1]
            else:
                self.snapshotting_cluster_id = self.member_clusters[0]
        self.cluster_enabled = True if self.cluster_mode == "enabled" else False
        self.cache_node_type = cache_node_type or "cache.t4g.micro"
        self.auth_token_enabled = bool(auth_token)
        self.auth_token_last_modified_date = (
            utcnow() if self.auth_token_enabled else None
        )
        self.transit_encryption_enabled = transit_encryption_enabled or False
        self.at_rest_encryption_enabled = at_rest_encryption_enabled or False
        self.kms_key_id = kms_key_id
        self.arn = f"arn:aws:elasticache:{region_name}:{account_id}:replicationgroup:{replication_group_id}"
        self.user_group_ids = user_group_ids or []
        self.log_delivery_configurations = self._get_log_delivery_configurations(
            log_delivery_configurations
        )
        self.replication_group_create_time = utcnow()
        self.data_tiering = "enabled" if data_tiering_enabled else "disabled"
        self.auto_minor_version_upgrade = auto_minor_version_upgrade
        self.network_type = network_type or "ipv4"
        self.ip_discovery = ip_discovery or "ipv4"
        self.transit_encryption_mode = transit_encryption_mode
        self.engine = engine or "redis"

        # Parameters not returned by create or describe replication group
        self.cache_security_group_names = cache_security_group_names
        self.cache_subnet_group_name = cache_subnet_group_name
        self.security_group_ids = security_group_ids
        self.tags = tags
        self.preferred_maintenance_window = preferred_maintenance_window
        self.notification_topic_arn = notification_topic_arn
        self.serverless_cache_snapshot_name = serverless_cache_snapshot_name
        self.cache_parameter_group_name = cache_parameter_group_name
        self.engine_version = engine_version
        self.snapshot_arns = snapshot_arns
        self.snapshot_name = snapshot_name

    @property
    def global_replication_group_info(self) -> dict[str, str]:
        return {
            "GlobalReplicationGroupId": self.global_replication_group_id,
            "GlobalReplicationGroupMemberRole": self.global_replication_group_member_role,
        }

    def _get_log_delivery_configurations(
        self, log_delivery_configurations: list[dict[str, Any]]
    ) -> list[dict[str, Any]]:
        log_delivery_configurations_resp = []
        if log_delivery_configurations:
            for log_delivery_configuration in log_delivery_configurations:
                log_config_resp = copy.copy(log_delivery_configuration)
                log_config_resp["status"] = "active"
                log_config_resp["message"] = "Log delivery configuration is active."

                log_delivery_configurations_resp.append(log_config_resp)
        return log_delivery_configurations_resp

    # This method will set populate the node_group with node_group_members
    # and member_clusters with the cache_cluster_ids of the primary and replicas.
    def _set_node_members_clusters_enabled(
        self,
        member_clusters: list[str],
        replication_group_id: str,
        replicas_per_node_group: int,
        node_group_configuration: list[dict[str, Any]],
        node_group: dict[str, Any],
        member_clusters_outpost_arns: list[str],
    ) -> None:
        replica_count = replicas_per_node_group
        primary_node = {}
        primary_node["cache_cluster_id"] = (
            f"{replication_group_id}-{node_group['node_group_id']}-001"
        )
        primary_node["cache_node_id"] = node_group["node_group_id"]

        # Set the PreferredAvailabilityZone,
        # default to us-east-1a for moto only
        primary_node["preferred_availability_zone"] = "us-east-1a"

        # With node_group_configurations
        node_config = {}
        if node_group_configuration:
            for config in node_group_configuration:
                if node_group["node_group_id"] == config.get("NodeGroupId"):
                    node_config = config
                    # overwrite slots if part of node config
                    if config.get("Slots"):
                        node_group["slots"] = config.get("Slots")

                    replica_count_val = config.get("ReplicaCount")
                    if replica_count_val is not None:
                        replica_count = int(replica_count_val)
                    else:
                        replica_count = replicas_per_node_group

                    primary_node["preferred_availability_zone"] = config.get(
                        "PrimaryAvailabilityZone", "us-east-1a"
                    )
                    primary_outpost_arn = config.get("PrimaryOutpostArn", "")
                    if primary_outpost_arn:
                        primary_node["preferred_outpost_arn"] = primary_outpost_arn
                        member_clusters_outpost_arns.append(primary_outpost_arn)

        node_group["node_group_members"].append(primary_node)
        member_clusters.append(primary_node["cache_cluster_id"])

        #  Loop through the number of replicas per node group
        for r in range(1, replica_count + 1):
            replica_node = {}
            # r + 1 because primary is 001
            replica_node["cache_cluster_id"] = (
                f"{replication_group_id}-{node_group['node_group_id']}-{r + 1:0>3}"
            )
            replica_node["cache_node_id"] = node_group["node_group_id"]

            # Set the ReplicaAvailabiilityZone,
            # default to us-east-1b for moto only
            replica_node["preferred_availability_zone"] = "us-east-1b"

            if node_config:
                replica_az = node_config.get("ReplicaAvailabilityZones")
                if replica_az:
                    if len(replica_az) >= r:
                        replica_node["preferred_availability_zone"] = replica_az[r - 1]
                    elif len(replica_az) >= 1:
                        replica_node["preferred_availability_zone"] = replica_az[0]
                replica_outpost_arns = node_config.get("ReplicaOutpostArns", [])
                if replica_outpost_arns:
                    if len(replica_outpost_arns) >= r:
                        replica_node["preferred_outpost_arn"] = replica_outpost_arns[
                            r - 1
                        ]
                    elif len(replica_outpost_arns) >= 1:
                        replica_node["preferred_outpost_arn"] = replica_outpost_arns[0]

                    member_clusters_outpost_arns.append(
                        replica_node["preferred_outpost_arn"]
                    )

            node_group["node_group_members"].append(replica_node)
            member_clusters.append(replica_node["cache_cluster_id"])

    def _set_node_members_clusters_disabled(
        self,
        member_clusters: list[str],
        replication_group_id: str,
        replication_group_domain: str,
        node_group: dict[str, Any],
        preferred_cache_cluster_azs: list[str],
        num_cache_clusters: int,
        replicas_per_node_group: int,
    ) -> None:
        primary_node: dict[str, Any] = {}
        primary_node["cache_cluster_id"] = f"{replication_group_id}-001"
        primary_node["cache_node_id"] = node_group["node_group_id"]
        primary_node["read_endpoint"] = {}
        primary_node["read_endpoint"]["address"] = (
            f"{primary_node['cache_cluster_id']}.{replication_group_domain}"
        )
        primary_node["read_endpoint"]["port"] = self.port

        if preferred_cache_cluster_azs:
            primary_node["preferred_availability_zone"] = preferred_cache_cluster_azs[0]
        else:
            primary_node["preferred_availability_zone"] = "us-east-1a"

        primary_node["current_role"] = "primary"
        node_group["node_group_members"].append(primary_node)
        member_clusters.append(primary_node["cache_cluster_id"])

        if replicas_per_node_group:
            num_cache_clusters = replicas_per_node_group + 1
        else:
            num_cache_clusters = 1

        # Use num_cache_clusters when cluster_mode is disabled
        for r in range(1, num_cache_clusters):
            replica_node: dict[str, Any] = {}
            # r + 1 because primary is 001
            replica_node["cache_cluster_id"] = f"{replication_group_id}-{r + 1:0>3}"
            replica_node["cache_node_id"] = node_group["node_group_id"]
            replica_node["read_endpoint"] = {}
            replica_node["read_endpoint"]["address"] = (
                f"{replica_node['cache_cluster_id']}.{replication_group_domain}"
            )
            replica_node["read_endpoint"]["port"] = self.port

            if preferred_cache_cluster_azs:
                if len(preferred_cache_cluster_azs) > r:
                    replica_az = preferred_cache_cluster_azs[r]
                elif len(preferred_cache_cluster_azs) >= 2:
                    replica_az = preferred_cache_cluster_azs[1]
                else:
                    replica_az = preferred_cache_cluster_azs[0]
            else:
                replica_az = "us-east-1b"

            replica_node["preferred_availability_zone"] = replica_az
            replica_node["current_role"] = "replica"

            node_group["node_group_members"].append(replica_node)
            member_clusters.append(replica_node["cache_cluster_id"])


class Snapshot(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        automatic_failover: Optional[bool],
        auto_minor_version_upgrade: Optional[bool],
        cache_cluster_create_time: Optional[str],
        cache_cluster_id: Optional[str],
        cache_node_type: Optional[str],
        cache_parameter_group_name: Optional[str],
        cache_subnet_group_name: Optional[str],
        engine: Optional[str],
        engine_version: Optional[str],
        kms_key_id: Optional[str],
        num_cache_nodes: Optional[int],
        num_node_groups: Optional[int],
        port: Optional[int],
        preferred_availability_zone: Optional[str],
        preferred_maintenance_window: Optional[str],
        preferred_outpost_arn: Optional[str],
        replication_group_description: Optional[str],
        replication_group_id: Optional[str],
        snapshot_name: Optional[str],
        snapshot_retention_limit: Optional[int],
        snapshot_source: Optional[str],
        snapshot_status: Optional[str],
        snapshot_window: Optional[str],
        topic_arn: Optional[str],
        tags: Optional[list[dict[str, str]]],
        vpc_id: Optional[str] = None,
    ):
        self.arn = f"arn:{get_partition(region_name)}:elasticache:{region_name}:{account_id}:snapshot:{snapshot_name}"
        self.automatic_failover = automatic_failover
        self.auto_minor_version_upgrade = auto_minor_version_upgrade
        self.cache_cluster_create_time = cache_cluster_create_time
        self.cache_cluster_id = cache_cluster_id
        self.cache_node_type = cache_node_type
        self.cache_parameter_group_name = cache_parameter_group_name
        self.cache_subnet_group_name = cache_subnet_group_name
        self.engine = engine
        self.engine_version = engine_version
        self.kms_key_id = kms_key_id
        self.num_cache_nodes = num_cache_nodes
        self.num_node_groups = num_node_groups
        self.port = port
        self.preferred_availability_zone = preferred_availability_zone
        self.preferred_maintenance_window = preferred_maintenance_window
        self.preferred_outpost_arn = preferred_outpost_arn
        self.replication_group_description = replication_group_description
        self.replication_group_id = replication_group_id
        self.snapshot_name = snapshot_name
        self.snapshot_retention_limit = snapshot_retention_limit
        self.snapshot_source = snapshot_source or "manual"
        self.snapshot_status = snapshot_status
        self.snapshot_window = snapshot_window
        self.topic_arn = topic_arn
        self.tags = tags or []
        self.vpc_id = vpc_id


class ElastiCacheBackend(BaseBackend):
    """Implementation of ElastiCache APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.arn_regex = re_compile(
            r"^arn:aws:elasticache:.*:[0-9]*:(cluster|snapshot|subnetgroup|replicationgroup|user):.*$"
        )
        self.users = {}
        self.users["default"] = User(
            account_id=self.account_id,
            region=self.region_name,
            user_id="default",
            user_name="default",
            engine="redis",
            access_string="on ~* +@all",
            no_password_required=True,
        )

        self.cache_clusters: dict[str, Any] = {}
        self.cache_subnet_groups: dict[str, CacheSubnetGroup] = {}
        self.replication_groups: dict[str, ReplicationGroup] = {}
        self.snapshots: dict[str, Snapshot] = {}
        self.tagging_service = TaggingService()

    def _get_snapshots_by_param(
        self,
        ref_id: Optional[str] = None,
        source: Optional[str] = None,
    ) -> list[Snapshot]:
        source_map = {
            "system": "automated",
            "user": "manual",
        }

        snapshots = list(self.snapshots.values())

        if ref_id:
            snapshots = [
                s
                for s in snapshots
                if s.cache_cluster_id == ref_id or s.replication_group_id == ref_id
            ]

        if source:
            actual_source = source_map.get(source.lower(), source.lower())
            snapshots = [s for s in snapshots if s.snapshot_source == actual_source]

        return snapshots

    def _get_vpc_id_from_cluster(self, cache_cluster_id: str) -> Optional[str]:
        if cache_cluster_id and cache_cluster_id in self.cache_clusters:
            cache_cluster = self.cache_clusters[cache_cluster_id]
            subnet_group_name = cache_cluster.cache_subnet_group_name
            if subnet_group_name and subnet_group_name in self.cache_subnet_groups:
                return self.cache_subnet_groups[subnet_group_name].vpc_id
        return None

    def create_user(
        self,
        user_id: str,
        user_name: str,
        engine: str,
        passwords: list[str],
        access_string: str,
        no_password_required: bool,
        authentication_type: str,  # contain it to the str in the enums TODO
        tags: Optional[list[dict[str, str]]] = None,
    ) -> User:
        if user_id in self.users:
            raise UserAlreadyExists

        if authentication_type not in AuthenticationTypes._value2member_map_:
            raise InvalidParameterValueException(
                f"Input Authentication type: {authentication_type} is not in the allowed list: [password,no-password-required,iam]"
            )

        if (
            no_password_required
            and authentication_type != AuthenticationTypes.NOPASSWORD
        ):
            raise InvalidParameterCombinationException(
                f"No password required flag is true but provided authentication type is {authentication_type}"
            )

        if passwords and authentication_type != AuthenticationTypes.PASSWORD:
            raise InvalidParameterCombinationException(
                f"Password field is not allowed with authentication type: {authentication_type}"
            )

        if not passwords and authentication_type == AuthenticationTypes.PASSWORD:
            raise InvalidParameterCombinationException(
                "A user with Authentication Mode: password, must have at least one password"
            )

        user = User(
            account_id=self.account_id,
            region=self.region_name,
            user_id=user_id,
            user_name=user_name,
            engine=engine,
            passwords=passwords,
            access_string=access_string,
            no_password_required=no_password_required,
            authentication_type=authentication_type,
            tags=tags,
        )

        if tags:
            self.tagging_service.tag_resource(user.arn, tags)

        self.users[user_id] = user
        return user

    def delete_user(self, user_id: str) -> User:
        if user_id in self.users:
            user = self.users[user_id]
            if user.status == "active":
                user.status = "deleting"
            return user
        raise UserNotFound(user_id)

    def describe_users(self, user_id: Optional[str]) -> list[User]:
        """
        Only the `user_id` parameter is currently supported.
        Pagination is not yet implemented.
        """
        if user_id:
            if user_id in self.users:
                user = self.users[user_id]
                if user.status == "deleting":
                    self.users.pop(user_id)
                return [user]
            else:
                raise UserNotFound(user_id)
        return list(self.users.values())

    def create_cache_cluster(
        self,
        cache_cluster_id: str,
        replication_group_id: str,
        az_mode: str,
        preferred_availability_zone: str,
        num_cache_nodes: int,
        cache_node_type: str,
        engine: str,
        engine_version: str,
        cache_parameter_group_name: str,
        cache_subnet_group_name: str,
        transit_encryption_enabled: bool,
        network_type: str,
        ip_discovery: str,
        snapshot_name: str,
        preferred_maintenance_window: str,
        port: int,
        notification_topic_arn: str,
        auto_minor_version_upgrade: bool,
        snapshot_retention_limit: int,
        snapshot_window: str,
        auth_token: str,
        outpost_mode: str,
        preferred_outpost_arn: str,
        preferred_availability_zones: list[str],
        cache_security_group_names: list[str],
        security_group_ids: list[str],
        tags: list[dict[str, str]],
        snapshot_arns: list[str],
        preferred_outpost_arns: list[str],
        log_delivery_configurations: list[dict[str, Any]],
        cache_node_ids_to_remove: list[str],
        cache_node_ids_to_reboot: list[str],
    ) -> CacheCluster:
        if cache_cluster_id in self.cache_clusters:
            raise CacheClusterAlreadyExists(cache_cluster_id)
        if cache_node_type is None or cache_node_type == "":
            raise InvalidParameterValueException(
                "The parameter CacheNodeType must be provided and must not be null."
            )
        cache_cluster = CacheCluster(
            account_id=self.account_id,
            region_name=self.region_name,
            cache_cluster_id=cache_cluster_id,
            replication_group_id=replication_group_id,
            az_mode=az_mode,
            preferred_availability_zone=preferred_availability_zone,
            preferred_availability_zones=preferred_availability_zones,
            num_cache_nodes=num_cache_nodes,
            cache_node_type=cache_node_type,
            engine=engine,
            engine_version=engine_version,
            cache_parameter_group_name=cache_parameter_group_name,
            cache_subnet_group_name=cache_subnet_group_name,
            cache_security_group_names=cache_security_group_names,
            security_group_ids=security_group_ids,
            tags=tags,
            snapshot_arns=snapshot_arns,
            snapshot_name=snapshot_name,
            preferred_maintenance_window=preferred_maintenance_window,
            port=port,
            notification_topic_arn=notification_topic_arn,
            auto_minor_version_upgrade=auto_minor_version_upgrade,
            snapshot_retention_limit=snapshot_retention_limit,
            snapshot_window=snapshot_window,
            auth_token=auth_token,
            outpost_mode=outpost_mode,
            preferred_outpost_arn=preferred_outpost_arn,
            preferred_outpost_arns=preferred_outpost_arns,
            log_delivery_configurations=log_delivery_configurations,
            transit_encryption_enabled=transit_encryption_enabled,
            network_type=network_type,
            ip_discovery=ip_discovery,
            cache_node_ids_to_remove=cache_node_ids_to_remove,
            cache_node_ids_to_reboot=cache_node_ids_to_reboot,
        )

        if tags:
            self.tagging_service.tag_resource(cache_cluster.arn, tags)

        self.cache_clusters[cache_cluster_id] = cache_cluster
        return cache_cluster

    def describe_cache_clusters(
        self, cache_cluster_id: Optional[str] = None
    ) -> list[CacheCluster]:
        if cache_cluster_id:
            if cache_cluster_id in self.cache_clusters:
                cache_cluster = self.cache_clusters[cache_cluster_id]
                return [cache_cluster]
            else:
                raise CacheClusterNotFound(cache_cluster_id)
        return list(self.cache_clusters.values())

    def delete_cache_cluster(self, cache_cluster_id: str) -> CacheCluster:
        if cache_cluster_id:
            if cache_cluster_id in self.cache_clusters:
                cache_cluster = self.cache_clusters[cache_cluster_id]
                cache_cluster.cache_cluster_status = "deleting"
                return cache_cluster
        raise CacheClusterNotFound(cache_cluster_id)

    def create_cache_subnet_group(
        self,
        cache_subnet_group_name: str,
        cache_subnet_group_description: str,
        subnet_ids: list[str],
        tags: Optional[list[dict[str, str]]],
    ) -> CacheSubnetGroup:
        if cache_subnet_group_name in self.cache_subnet_groups:
            raise CacheSubnetGroupAlreadyExists(cache_subnet_group_name)

        cache_subnet_group = CacheSubnetGroup(
            account_id=self.account_id,
            region_name=self.region_name,
            cache_subnet_group_name=cache_subnet_group_name,
            cache_subnet_group_description=cache_subnet_group_description,
            subnet_ids=subnet_ids,
            tags=tags,
        )

        if tags:
            self.tagging_service.tag_resource(cache_subnet_group.arn, tags)

        self.cache_subnet_groups[cache_subnet_group_name] = cache_subnet_group
        return cache_subnet_group

    def describe_cache_subnet_groups(
        self,
        cache_subnet_group_name: Optional[str] = None,
    ) -> list[CacheSubnetGroup]:
        if cache_subnet_group_name:
            if cache_subnet_group_name in self.cache_subnet_groups:
                cache_subnet_group = self.cache_subnet_groups[cache_subnet_group_name]
                return [cache_subnet_group]
            else:
                raise CacheSubnetGroupNotFound(cache_subnet_group_name)
        return list(self.cache_subnet_groups.values())

    def list_tags_for_resource(self, arn: str) -> dict[str, list[dict[str, str]]]:
        if self.arn_regex.match(arn):
            return self.tagging_service.list_tags_for_resource(arn)
        else:
            raise InvalidARNFault(arn)

    def add_tags_to_resource(self, arn: str, tags: list[dict[str, str]]) -> None:
        # Docs user "ResourceName" as input but the param is technically the ARN, will just use arn
        self.tagging_service.tag_resource(arn, tags)

    def remove_tags_from_resource(self, arn: str, tags: list[str]) -> None:
        # Docs user "ResourceName" as input but the param is technically the ARN, will just use arn
        self.tagging_service.untag_resource_using_names(arn, tags)

    def create_replication_group(
        self,
        replication_group_id: str,
        replication_group_description: str,
        global_replication_group_id: str,
        primary_cluster_id: str,
        automatic_failover_enabled: bool,
        multi_az_enabled: bool,
        num_cache_clusters: int,
        preferred_cache_cluster_azs: list[str],
        num_node_groups: int,
        replicas_per_node_group: int,
        node_group_configuration: list[dict[str, Any]],
        cache_node_type: str,
        engine: str,
        engine_version: str,
        cache_parameter_group_name: str,
        cache_subnet_group_name: str,
        cache_security_group_names: list[str],
        security_group_ids: list[str],
        tags: list[dict[str, str]],
        snapshot_arns: list[str],
        snapshot_name: str,
        preferred_maintenance_window: str,
        port: int,
        notification_topic_arn: str,
        auto_minor_version_upgrade: bool,
        snapshot_retention_limit: int,
        snapshot_window: str,
        auth_token: str,
        transit_encryption_enabled: bool,
        at_rest_encryption_enabled: bool,
        kms_key_id: str,
        user_group_ids: list[str],
        log_delivery_configurations: list[dict[str, Any]],
        data_tiering_enabled: bool,
        network_type: str,
        ip_discovery: str,
        transit_encryption_mode: str,
        cluster_mode: str,
        serverless_cache_snapshot_name: str,
    ) -> ReplicationGroup:
        if replication_group_id in self.replication_groups:
            raise ReplicationGroupAlreadyExists(replication_group_id)
        replication_group = ReplicationGroup(
            account_id=self.account_id,
            region_name=self.region_name,
            replication_group_id=replication_group_id,
            replication_group_description=replication_group_description,
            global_replication_group_id=global_replication_group_id,
            primary_cluster_id=primary_cluster_id,
            automatic_failover_enabled=automatic_failover_enabled,
            multi_az_enabled=multi_az_enabled,
            num_cache_clusters=num_cache_clusters,
            preferred_cache_cluster_azs=preferred_cache_cluster_azs,
            num_node_groups=num_node_groups,
            replicas_per_node_group=replicas_per_node_group,
            node_group_configuration=node_group_configuration,
            cache_node_type=cache_node_type,
            engine=engine,
            engine_version=engine_version,
            cache_parameter_group_name=cache_parameter_group_name,
            cache_subnet_group_name=cache_subnet_group_name,
            cache_security_group_names=cache_security_group_names,
            security_group_ids=security_group_ids,
            tags=tags,
            snapshot_arns=snapshot_arns,
            snapshot_name=snapshot_name,
            preferred_maintenance_window=preferred_maintenance_window,
            port=port,
            notification_topic_arn=notification_topic_arn,
            auto_minor_version_upgrade=auto_minor_version_upgrade,
            snapshot_retention_limit=snapshot_retention_limit,
            snapshot_window=snapshot_window,
            auth_token=auth_token,
            transit_encryption_enabled=transit_encryption_enabled,
            at_rest_encryption_enabled=at_rest_encryption_enabled,
            kms_key_id=kms_key_id,
            user_group_ids=user_group_ids,
            log_delivery_configurations=log_delivery_configurations,
            data_tiering_enabled=data_tiering_enabled,
            network_type=network_type,
            ip_discovery=ip_discovery,
            transit_encryption_mode=transit_encryption_mode,
            cluster_mode=cluster_mode,
            serverless_cache_snapshot_name=serverless_cache_snapshot_name,
        )

        if tags:
            self.tagging_service.tag_resource(replication_group.arn, tags)

        self.replication_groups[replication_group_id] = replication_group
        return replication_group

    def describe_replication_groups(
        self,
        replication_group_id: Optional[str] = None,
    ) -> list[ReplicationGroup]:
        if replication_group_id:
            if replication_group_id in self.replication_groups:
                replication_group = self.replication_groups[replication_group_id]
                return [replication_group]
            else:
                raise ReplicationGroupNotFound(replication_group_id)
        return list(self.replication_groups.values())

    def delete_replication_group(
        self, replication_group_id: str, retain_primary_cluster: Optional[bool]
    ) -> ReplicationGroup:
        if replication_group_id not in self.replication_groups:
            raise ReplicationGroupNotFound(replication_group_id)

        replication_group = self.replication_groups[replication_group_id]
        replication_group.status = "deleting"
        if not retain_primary_cluster:
            replication_group = self.replication_groups[replication_group_id]
            primary_id = replication_group.primary_cluster_id
            if primary_id and primary_id in replication_group.member_clusters:
                replication_group.member_clusters.remove(primary_id)
            replication_group.primary_cluster_id = ""
        return replication_group

    def create_snapshot(
        self,
        snapshot_name: str,
        replication_group_id: Optional[str],
        cache_cluster_id: Optional[str],
        kms_key_id: Optional[str],
        tags: Optional[list[dict[str, str]]],
    ) -> Snapshot:
        resource = None
        num_node_groups = None
        vpc_id = None

        if snapshot_name in self.snapshots:
            raise SnapshotAlreadyExists(snapshot_name)
        if replication_group_id and cache_cluster_id:
            raise InvalidParameterCombinationException(
                "Only one of CacheClusterId or ReplicationGroupId can be provided."
            )
        if replication_group_id:
            if replication_group_id not in self.replication_groups:
                raise ReplicationGroupNotFound(replication_group_id)
            resource = self.replication_groups[replication_group_id]
            num_node_groups = len(resource.node_groups)
        elif cache_cluster_id:
            if cache_cluster_id not in self.cache_clusters:
                raise CacheClusterNotFound(cache_cluster_id)
            resource = self.cache_clusters[cache_cluster_id]
            vpc_id = self._get_vpc_id_from_cluster(cache_cluster_id)
        else:
            raise InvalidParameterValueException(
                "One of CacheClusterId or ReplicationGroupId must be provided."
            )

        snapshot = Snapshot(
            account_id=self.account_id,
            region_name=self.region_name,
            automatic_failover=getattr(resource, "automatic_failover", None),
            auto_minor_version_upgrade=getattr(
                resource, "auto_minor_version_upgrade", None
            ),
            cache_cluster_create_time=getattr(
                resource, "cache_cluster_create_time", None
            ),
            cache_cluster_id=getattr(resource, "cache_cluster_id", None),
            cache_node_type=getattr(resource, "cache_node_type", None),
            cache_parameter_group_name=getattr(
                resource, "cache_parameter_group_name", None
            ),
            cache_subnet_group_name=getattr(resource, "cache_subnet_group_name", None),
            engine=getattr(resource, "engine", None),
            engine_version=getattr(resource, "engine_version", None),
            kms_key_id=kms_key_id,
            num_cache_nodes=getattr(resource, "num_cache_nodes", None),
            num_node_groups=num_node_groups,
            port=getattr(resource, "port", None),
            preferred_availability_zone=getattr(
                resource, "preferred_availability_zone", None
            ),
            preferred_maintenance_window=getattr(
                resource, "preferred_maintenance_window", None
            ),
            preferred_outpost_arn=getattr(resource, "preferred_outpost_arn", None),
            replication_group_description=getattr(resource, "description", None),
            replication_group_id=getattr(resource, "replication_group_id", None),
            snapshot_name=snapshot_name,
            snapshot_retention_limit=getattr(
                resource, "snapshot_retention_limit", None
            ),
            snapshot_status="available",
            snapshot_source="manual",
            snapshot_window=getattr(resource, "snapshot_window", None),
            topic_arn=getattr(resource, "notification_topic_arn", None),
            tags=tags or [],
            vpc_id=vpc_id,
        )

        if tags:
            self.tagging_service.tag_resource(snapshot.arn, tags)

        self.snapshots[snapshot_name] = snapshot
        return snapshot

    def describe_snapshots(
        self,
        replication_group_id: Optional[str] = None,
        cache_cluster_id: Optional[str] = None,
        snapshot_name: Optional[str] = None,
        snapshot_source: Optional[str] = None,
        marker: Optional[str] = None,
        max_records: Optional[int] = None,
        show_node_group_config: Optional[bool] = None,
    ) -> list[Snapshot]:
        if snapshot_name:
            try:
                return [self.snapshots[snapshot_name]]
            except KeyError:
                raise SnapshotNotFound(snapshot_name)

        if replication_group_id and cache_cluster_id:
            raise InvalidParameterCombinationException(
                "Only one of CacheClusterId or ReplicationGroupId can be provided."
            )

        if replication_group_id and replication_group_id not in self.replication_groups:
            raise ReplicationGroupNotFound(replication_group_id)

        if cache_cluster_id and cache_cluster_id not in self.cache_clusters:
            raise CacheClusterNotFound(cache_cluster_id)

        ref_id = replication_group_id or cache_cluster_id

        if snapshot_source or ref_id:
            return self._get_snapshots_by_param(ref_id=ref_id, source=snapshot_source)

        return list(self.snapshots.values())

    def delete_snapshot(self, snapshot_name: str) -> Snapshot:
        if snapshot_name in self.snapshots:
            snapshot = self.snapshots[snapshot_name]
            snapshot.snapshot_status = "deleting"

            self.snapshots.pop(snapshot_name)

            return snapshot
        raise SnapshotNotFound(snapshot_name)


elasticache_backends = BackendDict(ElastiCacheBackend, "elasticache")



"""
Response handlers for AWS CloudWatch Synthetics API emulation in Moto.
"""

import json

from moto.core.responses import BaseResponse
from moto.synthetics.models import SyntheticsBackend, synthetics_backends


class SyntheticsResponse(BaseResponse):
    """
    Handles API responses for AWS CloudWatch Synthetics operations.
    """

    def __init__(self) -> None:
        """
        Initialize the SyntheticsResponse with the synthetics service name.
        """
        super().__init__(service_name="synthetics")

    @property
    def synthetics_backend(self) -> SyntheticsBackend:
        """
        Returns the backend instance for the current region.
        """
        return synthetics_backends[self.current_account][self.region]

    def create_canary(self) -> str:
        """
        Create a new canary using the provided parameters.
        """
        params = json.loads(self.body)
        canary = self.synthetics_backend.create_canary(
            name=params["Name"],
            code=params.get("Code", {}),
            artifact_s3_location=params.get("ArtifactS3Location", "s3://dummy"),
            execution_role_arn=params.get(
                "ExecutionRoleArn", "arn:aws:iam::123:role/service-role"
            ),
            schedule=params.get("Schedule", {"Expression": "rate(5 minutes)"}),
            run_config=params.get("RunConfig", {"TimeoutInSeconds": 60}),
            success_retention_period_in_days=params.get(
                "SuccessRetentionPeriodInDays", 31
            ),
            failure_retention_period_in_days=params.get(
                "FailureRetentionPeriodInDays", 31
            ),
            runtime_version=params.get("RuntimeVersion", "syn-nodejs-puppeteer-3.8"),
            vpc_config=params.get("VpcConfig"),
            resources_to_replicate_tags=params.get("ResourcesToReplicateTags"),
            provisioned_resource_cleanup=params.get("ProvisionedResourceCleanup"),
            browser_configs=params.get("BrowserConfigs"),
            tags=params.get("Tags", {}),
            artifact_config=params.get("ArtifactConfig"),
        )
        return json.dumps({"Canary": canary.to_dict()})

    def get_canary(self) -> str:
        """
        Retrieve details for a specific canary by name.
        """
        # Extract name from the URL path /canary/MyCanary
        path_parts = self.path.split("/")
        name = path_parts[-1] if len(path_parts) > 1 else None
        if not name:
            raise ValueError("Canary name not found in URL")
        canary = self.synthetics_backend.get_canary(name)
        return json.dumps({"Canary": canary.to_dict()})

    def describe_canaries(self) -> str:
        """
        List all canaries in the backend.
        """
        canaries, _ = self.synthetics_backend.describe_canaries(None, None, None)
        return json.dumps({"Canaries": [c.to_dict() for c in canaries]})

    def list_tags_for_resource(self) -> str:
        """
        List tags for a given resource ARN.
        """
        # Extract ARN from the URL path /tags/{resourceArn}
        path_parts = self.path.split("/")
        arn = path_parts[-1] if len(path_parts) > 1 else None
        if not arn:
            raise ValueError("Resource ARN not found in URL")
        tags = self.synthetics_backend.list_tags_for_resource(arn)
        return json.dumps({"Tags": tags})

    def _get_action(self) -> str:
        """
        Override to provide default action for root endpoint.
        """
        action = super()._get_action()
        if action is None and self.path == "/":
            return "GetHealthCheck"  # Default action for root endpoint
        return action or "Unknown"

    def get_health_check(self) -> str:
        """
        Handle root endpoint requests.
        """
        return "What would you like to do?"



import abc
from typing import Any, Final, Optional

from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringVariableSample,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class Argument(EvalComponent, abc.ABC):
    """
    Represents an Intrinsic Function argument that can be evaluated and whose
    result is pushed onto the stack.

    Subclasses must override `_eval_argument()` to evaluate the specific value
    of the argument they represent. This abstract class manages the type and
    environment handling by appending the evaluated result to the environment's
    stack in `_eval_body`.

    The `_eval_body` method calls `_eval_argument()` and pushes the resulting
    value to the stack.
    """

    @abc.abstractmethod
    def _eval_argument(self, env: Environment) -> Any: ...

    def _eval_body(self, env: Environment) -> None:
        argument = self._eval_argument(env=env)
        env.stack.append(argument)


class ArgumentLiteral(Argument):
    definition_value: Final[Optional[Any]]

    def __init__(self, definition_value: Optional[Any]):
        self.definition_value = definition_value

    def _eval_argument(self, env: Environment) -> Any:
        return self.definition_value


class ArgumentJsonPath(Argument):
    json_path: Final[str]

    def __init__(self, json_path: str):
        self.json_path = json_path

    def _eval_argument(self, env: Environment) -> Any:
        inp = env.stack[-1]
        value = extract_json(self.json_path, inp)
        return value


class ArgumentContextPath(ArgumentJsonPath):
    def __init__(self, context_path: str):
        json_path = context_path[1:]
        super().__init__(json_path=json_path)

    def _eval_argument(self, env: Environment) -> Any:
        value = extract_json(
            self.json_path, env.states.context_object.context_object_data
        )
        return value


class ArgumentFunction(Argument):
    function: Final[EvalComponent]

    def __init__(self, function: EvalComponent):
        self.function = function

    def _eval_argument(self, env: Environment) -> Any:
        self.function.eval(env=env)
        output_value = env.stack.pop()
        return output_value


class ArgumentVar(Argument):
    string_variable_sample: Final[StringVariableSample]

    def __init__(self, string_variable_sample: StringVariableSample):
        super().__init__()
        self.string_variable_sample = string_variable_sample

    def _eval_argument(self, env: Environment) -> Any:
        self.string_variable_sample.eval(env=env)
        value = env.stack.pop()
        return value


class ArgumentList(Argument):
    arguments: Final[list[Argument]]
    size: Final[int]

    def __init__(self, arguments: list[Argument]):
        self.arguments = arguments
        self.size = len(arguments)

    def _eval_argument(self, env: Environment) -> Any:
        values = []
        for argument in self.arguments:
            argument.eval(env=env)
            argument_value = env.stack.pop()
            values.append(argument_value)
        return values
